# -*- coding: utf-8 -*-
"""
Last modified on Sep 06, 2024

@author: Hermann Zeyen <hermann.zeyen@universite-paris-saclay.fr>
         University Paris-Saclay, France

Contains functions for data input/output

Contains methods:
    get_files
    read_geography_file
    get_mag_field

Contains class data with the following functions:
    __init__
    read_geometrics
    read_txt
    read_gxf
    read_BRGM_flight
    get_line
    lines
    write_dat
    clean_data
    line_days
    get_segment
    julian2date
    date2julian
    read_base
    write_base
    diurnal_correction
    diurnal_variation
    interpol_line
    interpol_2D
    extrapolate
    justify_lines_median
    justify_lines_gaussian
    matrixExtension
    poleReduction
    plot_geography
    plot_triang
    plot_image
    addMPL
    rmMPL

"""

#        fill_nans (actually not used)
#            eliminate_nans (actually not used)
import sys
import os
import numpy as np
from PyQt5 import QtWidgets

# from ..plotting import plot as w
from .dialog import dialog
from .geometrics import Geometrics


def get_files(dir0=None, ftype=None):
    """
    Ask for files with ending "stn".
    Several files may be chosen at once using as usual SHFT or CTRL.
    If a folder was chosen, it is automatically recognized and eliminated
    from the list of file names.

    Returns
    -------
    data_files: list str
        list of chosen files

    """
    valid_extensions = np.array([".STN", ".OUT", ".XYZ", ".GXF", ".DAT"])
    ftypes = ("GEOMETRICS", "MGWIN", "BRGM", "GXF", "OTHER")
    dtypes = ("magnetic", "gravity")
    #    sensor2_types = ["GEOMETRICS", "MGWIN"]
    try:
        os.chdir(dir0)
    except (FileNotFoundError, TypeError):
        dir0 = None
    files = list(
        QtWidgets.QFileDialog.getOpenFileNames(
            None,
            "Select data files",
            "",
            filter="stn/gxf/XYZ/dat (*.stn *.gxf *.XYZ *.dat) ;; all (*.*)",
        )
    )
    if len(files) == 0:
        print("No file chosen, program finishes")
        sys.exit("No file chosen")
    elif len(files[0]) == 0:
        print(
            "\nNo file chosen, program finishes\n\n"
            + "You probably must close the Spyder console before restarting"
        )
        sys.exit("No file chosen")
    # Sort chosen file names
    files[0].sort()
    # Check data formats
    # Set working folder as folder of the first selected file
    if ftype != "base":
        dir0 = os.path.dirname(files[0][0])
        os.chdir(dir0)
        print(f"Set working folder to: {dir0}")

    # Loop over file names and store valid file names into list data_files
    data_files = []
    file_types = []
    for _, f in enumerate(files[0]):
        _, file_ext = os.path.splitext(f)
        file_ext = file_ext.upper()
        if os.path.isdir(f):
            continue
        if file_ext not in valid_extensions:
            continue
        file_types.append(ftypes[np.where(valid_extensions == file_ext)[0][0]])
        data_files.append(f)
    if len(data_files) == 0:
        _ = QtWidgets.QMessageBox.critical(
            None,
            "Error",
            "No valid data files given\n\n"
            + f"Only {valid_extensions} allowed.\n\nProgram stops",
            QtWidgets.QMessageBox.Ok,
        )
        raise Exception("File type error.\n")
    # Ask for data types
    labels = []
    values = []
    types = []
    if ftype == "base":
        data_types = []
        for _ in range(len(files[0])):
            data_types.append("magnetic")
    else:
        for f in data_files:
            labels.append(f"{os.path.basename(f)}:")
            values.append(None)
            types.append("l")
            labels.append(["Magnetic", "Gravity"])
            values.append("0")
            types.append("r")
        results, ok_button = dialog(labels, types, values, title="data types")
        if not ok_button:
            print("No entry, program finished")
            sys.exit()
        data_types = []
        for i in range(1, len(results), 2):
            data_types.append(dtypes[int(results[i])])
    return data_files, file_types, data_types, dir0


def read_geography_file(file):
    """
    Reads file with geography information to be plotted (borders - mainly geological, towns)

    Parameters
    ----------
    file : str, name of file to be read
        File has the following structure:

        - keyword may be "#POINT", "#LINE" or "#END"
        - if keyword == "#POINT", one line follows with x, y coordinates and text
          text being the description of the point (no blanks)
        - if keyword == "#LINE", several lines follow, each one with x and y
          coordinate of one point describing the line
        - if keyword == "#END", this line finishes the data entry, possible following
          lines will be ignored.

    Returns
    -------
    geography : dictionary with all geography information.
        key is consecutive numbering of entries
        Each entry consists of a dictionary with the following entries:

        - "type" str
          may be "POINT" or "LINE"

          - If type == "POINT" : One line with:

              - "x" : float: x coordinate of point (East)
              - "y" : float: y coordinate of point (North)
              - "name" : str: Text to be plotted beside the point mostly name of a town

          - If type == "line" :

              - "x" : list of floats, East coordinate of points describing the line
              - "y" : list of floats, North coordinate of points describing the line

    """
    with open(file, "r", encoding="utf-8") as fi:
        ll = fi.readlines()
    geography = {}
    il = 0
    iunit = -1
    while True:
        if ll[il].upper().startswith("#POINT"):
            iunit += 1
            il += 1
            nums = ll[il].split()
            geography[iunit] = {}
            geography[iunit]["type"] = "POINT"
            geography[iunit]["x"] = float(nums[0])
            geography[iunit]["y"] = float(nums[1])
            geography[iunit]["name"] = nums[2]
            il += 1
        elif ll[il].upper().startswith("#LINE"):
            iunit += 1
            geography[iunit] = {}
            geography[iunit]["type"] = "LINE"
            geography[iunit]["x"] = []
            geography[iunit]["y"] = []
            while True:
                il += 1
                if ll[il].startswith("#"):
                    break
                nums = ll[il].split()
                geography[iunit]["x"].append(float(nums[0]))
                geography[iunit]["y"].append(float(nums[1]))
        elif ll[il].upper().startswith("#END"):
            break
    return geography


def get_mag_field():
    """
    Get parameters of Earth's magnetic field in the study area

    Returns
    -------

    inclination : float
        Inclination of Earth's field in degrees
    declination : float
        Declination of Earth's field in degrees
    """
    results, _ = dialog(
        ["Field inclination", "Field declination"],
        ["e", "e"],
        [62, 0],
        "Earth's field parameters",
    )
    return float(results[0]), float(results[1])


class Data:
    """
    Class contains methods for data management in program PyMaGra
    """

    def __init__(self, n_block):
        """
        Initialisation of class Data

        Parameters
        ----------
        n_block : int
            Number of data set read.

        Returns
        -------
        None.

        """
        self.sensor1 = []
        self.sensor2 = []
        self.sensor1_ori = []
        self.sensor2_ori = []
        self.gdata = []
        self.base = []
        self.grad = np.array([])
        self.grad_ori = np.array([])
        self.x_inter = np.array([])
        self.y_inter = np.array([])
        self.x = []
        self.y = []
        self.z = []
        self.topo = []
        self.dispo = 0
        self.sensor1_inter = np.zeros((1, 1))
        self.sensor2_inter = np.zeros((1, 1))
        self.time = []
        self.segments = {}
        self.grad_data = False
        self.d_sensor = 0.9
        self.h_sensor = 0.4
        self.line_declination = 0.0
        self.interpol = 0.2
        self.data = {}
        self.n_block = n_block
        self.n_lines = 0
        self.n_data = 0
        self.direction = 0
        self.dx = 0.0
        self.dy = 0.0
        self.line_choice = "all"

    def read_geometrics(self, file):
        """
        Read Geometrics .stn file (G-858 instrument)

        Parameters
        ----------
        file : str
            Name of data file.

        Returns
        -------
        data :  Dictionary with keys equal to line numbers (starting at 0)
            Each line is itself a dictionary with the following entries:
            Key is line number

            - "s1": Numpy float array with data of sensor 1
            - "s2": Numpy float array with data of sensor 2
              If only data of one single sensor were measured, "s2" contains
              only one zero.
            - "x":  Numpy float array with E-W coordinates of data points
            - "y":  Numpy float array with N-S coordinates of data points
            - "grad_flag" bool. True if 2 sensors were used, False if only one sensor
            - "mask": bool, True if line should be plotted, False if excluded from
              plotting. Initially set to True

        The original data are stored in class geometrics.Geometrics. See file
        geometrics.py for documentation

        """
        labels = [
            "Height of sensor 1 above ground [m]",
            ["Vertical disposition", "Horizontal disposition"],
            "Distance between sensors [m]\n"
            + "  positive: sensor 2 above or right of sensor 1",
            "Direction of lines [degrees from N to E]",
        ]
        types = ["e", "r", "e", "e"]
        values = [self.h_sensor, self.dispo, self.d_sensor, self.line_declination]
        results, ok_button = dialog(
            labels, types, values, title="Geometrics parameters"
        )
        if not ok_button:
            print("Program aborted")
            sys.exit()
        self.h_sensor = float(results[0])
        self.dispo = int(results[1])
        self.d_sensor = float(results[2])
        self.line_declination = float(results[3])
        self.gdata.append(Geometrics())
        self.gdata[-1].read_stn(
            file, self.n_block, self.d_sensor, self.h_sensor, self.dispo
        )
        self.segments = self.gdata[-1].segments
        self.sensor1 = self.gdata[-1].sensor1
        self.sensor2 = self.gdata[-1].sensor2
        if len(self.sensor1) == len(self.sensor2):
            self.grad = (self.sensor2 - self.sensor1) / self.d_sensor
            self.grad_data = True
        else:
            self.grad = np.array([0.0])
            self.grad_data = False
        self.x = self.gdata[-1].x
        self.y = self.gdata[-1].y
        self.z = np.ones_like(self.x) * self.h_sensor
        self.topo = np.zeros_like(self.x)
        self.n_data = self.gdata[-1].n_data
        self.n_lines = self.gdata[-1].n_lines
        self.sensor1_ori = np.copy(self.sensor1)
        self.sensor2_ori = np.copy(self.sensor2)
        self.grad_ori = np.copy(self.grad)
        self.time = self.gdata[-1].time
        self.direction = 0
        self.dx = self.segments[0]["dx"][0]
        self.dy = self.segments[0]["dy"][0]
        self.data = self.lines()
        self.data["grad_data"] = self.grad_data
        self.data["year"] = self.gdata[-1].year[0]
        self.data["dispo"] = self.dispo
        self.data["block"] = self.n_block
        self.data["height"] = self.h_sensor
        self.data["line_declination"] = self.line_declination
        if self.grad_data:
            self.data["height2"] = self.h_sensor + self.d_sensor
            self.data["d_sensor"] = self.d_sensor
        return self.data

    def write_geometrics(self, file, data1, x, y, data2=None, n_block=0):
        """
        Wrapper to write data in Geometrics MagMap2000 .stn format.

        Data must be interpolated onto a regular grid.

        Parameters
        ----------
        file : str
            File name where to write data.
        data1 : numpy float array [number_of_samples_per_line, number_of_lines]
            Data of sensor 1.
        x : numpy float array [number_of_samples_per_line]
            X coordinates of all measured points.
        y : numpy float array [number_of_samples_per_line, number_of_lines]
            Y coordinates of all measured points.

        Optional parameters:

        data2 : numpy float array [number_of_samples_per_line, number_of_lines]
            Data of sensor 2. Optional. Default: np.zeros_like(data1)
        n_block : int, optional
            Number of block (data set) to be written. The default is 0.

        Returns
        -------
        None.

        """
        if not isinstance(data1, np.ndarray):
            n_block = data1["block"] - 1
        self.gdata[n_block].write_stn(file, data1, x, y, data2=data2)

    def read_txt(self, file):
        """
        Reads a non-Geometrics format magnetic data file
        This option is mainly thought for reading the output of program mgwin
        used with the option to enter all data points with their specific positions
        (NBPTS > 0). This allows calculation of a 2D map with mgwin. You may use
        Prepare_mgwin_calculation_points.py to preapre the coordinates.

        The file structure is as follows:

        - One comment line
        - line with NX, NY (number of points in X and Y direction)
          It is supposed that the data have been calculated on a regular
          grid. mgwin writes on this line the total number of data points
          the file must therefore be edited to replace the existing number
          by the two required ones.
        - one line per data poin with (X, Y, Z, DATA)

        mgwin writes only one value into the file and in order to keep the
        structure of Geometrics simple, these values are copied into both
        sensor1 and sensor2 arrays.  Data are copied as well into
        self.sensor_n (1D array) and self.sensor_n_inter (2D array)

        Parameters
        ----------
        file : str, name of file to be read

        Returns
        -------
        data

        """
        labels = [
            "Height of lower sensor above ground [m]",
            "Distance between sensors [m]",
            "Direction of lines [degrees from N to E]",
        ]
        types = ["e", "e", "e"]
        values = [self.h_sensor, self.d_sensor, self.line_declination]
        results, ok_button = dialog(labels, types, values, title="MGWin parameters")
        if not ok_button:
            print("Program aborted")
            sys.exit()
        self.h_sensor = float(results[0])
        self.d_sensor = float(results[1])
        self.line_declination = float(results[2])
        with open(file, "r", encoding="utf-8") as fi:
            lines = fi.readlines()
        nums = lines[1].split()
        nx = int(nums[0])
        ny = int(nums[1])
        self.sensor1_inter = np.zeros((ny, nx))
        xx = np.zeros((ny, nx))
        yy = np.zeros((ny, nx))
        zz = np.zeros((ny, nx))
        ll = lines[2].split()
        if len(ll) > 4:
            grad_flag = True
            self.grad_data = True
            self.sensor2_inter = np.zeros((ny, nx))
        # elif len(ll) <4:
        #     grad_flag = False
        else:
            grad_flag = False
        n = 1
        line = -1
        for i in range(nx):
            line += 1
            self.segments[line] = {}
            n1 = n - 1
            self.segments[line]["mark_samples"] = [n1]
            self.segments[line]["dx"] = []
            self.segments[line]["dy"] = []
            self.segments[line]["d"] = []
            for j in range(ny):
                n += 1
                nums = lines[n].split()
                xx[j, i] = float(nums[0])
                yy[j, i] = float(nums[1])
                zz[j, i] = float(nums[2])
                self.x.append(float(nums[0]))
                self.y.append(float(nums[1]))
                self.sensor1_inter[j, i] = float(nums[3])
                self.sensor1.append(float(nums[3]))
                if grad_flag:
                    self.sensor2_inter[j, i] = float(nums[4])
                    self.sensor2.append(float(nums[4]))
                # else:
                #     self.sensor2_inter[j,i] = self.sensor1_inter[j,i]
                #     self.sensor2.append(float(nums[3]))
            self.segments[line]["mark_samples"].append(n - 1)
            n2 = n - 2
            self.segments[line]["dx"].append(abs(self.x[-1] - self.x[-2]))
            self.segments[line]["dy"].append(abs(self.y[-1] - self.y[-2]))
            self.segments[line]["d"].append(
                np.sqrt(
                    self.segments[line]["dx"][-1] ** 2
                    + self.segments[line]["dy"][-1] ** 2
                )
            )
            self.segments[line]["median1"] = np.median(self.sensor1[n1:n2])
            self.segments[line]["median2"] = np.median(self.sensor2_inter[n1:n2])
            self.segments[line]["x"] = np.median(self.x[n1:n2])
            self.segments[line]["y"] = np.median(self.y[n1:n2])
            self.segments[line]["mask"] = True
            self.segments[line]["block"] = self.n_block
            self.segments[line]["direction"] = self.line_declination
            self.segments[line]["dir"] = "odd"
            self.segments[line]["pos"] = self.segments[line]["x"]
            if grad_flag:
                self.segments[line]["sensor"] = 0
            else:
                self.segments[line]["sensor"] = 1
        self.sensor1 = np.array(self.sensor1)
        self.sensor2 = np.array(self.sensor2)
        self.n_data = len(self.sensor1)
        self.n_lines = nx
        if grad_flag:
            self.grad = (self.sensor1 - self.sensor2) / self.d_sensor
        # Store original data to arrays xxx_ori
        self.sensor1_ori = np.copy(self.sensor1)
        self.sensor2_ori = np.copy(self.sensor2)
        self.grad_ori = np.copy(self.grad)
        self.x = np.array(self.x)
        self.y = np.array(self.y)
        self.z = np.ones_like(self.x) * self.h_sensor
        self.topo = np.zeros_like(self.x)
        self.time = np.arange(len(self.x)) * 0.1
        self.x_inter = np.unique(self.x)
        self.y_inter = np.unique(self.y)
        self.direction = 0
        self.dx = self.segments[0]["dx"][0]
        self.dy = self.segments[0]["dy"][0]
        self.data = self.lines()
        self.data["grad_data"] = grad_flag
        self.data["year"] = 0
        self.data["dispo"] = 0
        self.data["block"] = self.n_block
        self.data["height"] = self.h_sensor
        self.data["line_declination"] = self.line_declination
        if grad_flag:
            self.data["height2"] = self.h_sensor + self.d_sensor
        return self.data

    def read_gxf(self, infile):
        """
        Read a gxf file (BRGM magnetic and gravity gridded files)

        Parameters
        ----------
        infile: string
            Name of file to be read
        """
        labels = [
            "Height of sensor above ground [m]",
            "Direction of lines [degrees from N to E]",
        ]
        types = ["e", "e"]
        values = [85.0, 0.0]
        results, ok_button = dialog(labels, types, values, title="MGWin parameters")
        if not ok_button:
            print("Program aborted")
            sys.exit()
        self.h_sensor = float(results[0])
        self.line_declination = float(results[1])
        with open(infile, "r", encoding="utf-8") as fi:
            lines = fi.readlines()
        # Read header
        il = -1
        n_rows = 0
        n_cols = 0
        x_origin = 0.0
        y_origin = 0.0
        dummy = 0.0
        while True:
            il += 1
            if lines[il][:5] == "#GRID":
                break
            if lines[il][:7] == "#POINTS":
                il += 1
                n_cols = int(lines[il])
            elif lines[il][:5] == "#ROWS":
                il += 1
                n_rows = int(lines[il])
            elif lines[il][:13] == "#PTSEPARATION":
                il += 1
                self.dx = float(lines[il])
            elif lines[il][:13] == "#RWSEPARATION":
                il += 1
                self.dy = float(lines[il])
            elif lines[il][:8] == "#XORIGIN":
                il += 1
                x_origin = float(lines[il])
            elif lines[il][:8] == "#YORIGIN":
                il += 1
                y_origin = float(lines[il])
            elif lines[il][:6] == "#DUMMY":
                il += 1
                dummy = float(lines[il])
        data = np.zeros((n_rows, n_cols))
        self.n_lines = n_rows
        self.grad_data = False
        self.direction = 1
        self.x_inter = x_origin + np.arange(n_cols) * self.dx
        self.y_inter = y_origin + np.arange(n_rows) * self.dy
        # Read data
        c2 = 0
        for ir in range(n_rows):
            c1 = c2
            self.segments[ir] = {}
            ic2 = 0
            while True:
                il += 1
                line = np.array(lines[il].split(), dtype=float)
                ic1 = ic2
                ic2 += len(line)
                data[ir, ic1:ic2] = line
                if ic2 == n_cols:
                    break
            data[data == dummy] = np.nan
            d = data[ir, :]
            index = np.isfinite(d)
            d = d[index]
            c2 = c1 + len(d)
            self.sensor1.extend(list(d))
            self.x.extend(list(self.x_inter[index]))
            self.y.extend(list(np.ones_like(self.x_inter[index]) * self.y_inter[ir]))
            self.segments[ir]["mark_samples"] = [c1, c2]
            self.segments[ir]["dx"] = [self.dx]
            self.segments[ir]["dy"] = [0]
            self.segments[ir]["d"] = [self.dx]
            self.segments[ir]["median1"] = np.median(d)
            self.segments[ir]["median2"] = 0.
            self.segments[ir]["x"] = np.median(self.x_inter)
            self.segments[ir]["y"] = self.y_inter[ir]
            self.segments[ir]["mask"] = True
            self.segments[ir]["block"] = self.n_block
            self.segments[ir]["dir"] = "odd"
            self.segments[ir]["pos"] = self.y_inter[ir]
            self.segments[ir]["direction"] = self.line_declination
            self.segments[ir]["sensor"] = 1
        self.x = np.array(self.x)
        self.y = np.array(self.y)
        self.z = np.ones_like(self.x) * self.h_sensor
        self.topo = np.zeros_like(self.x)
        self.sensor1 = np.array(self.sensor1)
        # Store original data to arrays xxx_ori
        self.sensor1_ori = np.copy(self.sensor1)
        self.sensor1_inter = np.copy(data)
        self.time = np.arange(len(self.x))
        self.data = self.lines()
        self.data["grad_data"] = False
        self.data["year"] = 0
        self.data["dispo"] = 0
        self.data["block"] = self.n_block
        self.data["height"] = self.h_sensor
        self.data["line_declination"] = self.line_declination
        return self.data

    def read_BRGM_flight(self, file):
        """
        Reads magnetic data from flight lines out of a BRGM data file

        Parameters
        ----------
        file : str
            Name of file containing the data.

        Returns
        -------
        x :     1D numpy float array
            E coordinate of each measured point along the line [Lambert 2 extended, meters]
        y :     1D numpy float array
            N coordinate of each measured point along the line [Lambert 2 extended, meters]
        v :     1D numpy float array
            Magnetic anomaly (measured field minus IGRF) [nT]
        topo :  1D numpy float array
            DMT topography at each measured point along the line [m]
        height : 1D numpy float array
            Flight height above topo at each measured point along the line [m]
        num : int
            Number of flight line (same as line if line < 100000.)
        """
        line_number = []
        nl = -1
        x = []
        y = []
        t = []
        height = []
        v = []
        c2 = 0
        with open(file, "r", encoding="utf-8") as fi:
            lines = fi.readlines()
        for _, line in enumerate(lines):
            if "/" in line:
                continue
            if "Line" in line:
                line_number.append(int(line.split()[1]))
                nl1 = nl
                nl += 1
                if nl == 0:
                    continue
                self.sensor1.extend(v)
                self.x.extend(x)
                self.y.extend(y)
                self.z.extend(height)
                self.topo.extend(t)
                self.segments[nl1] = {}
                c1 = c2
                c2 += len(x)
                self.segments[nl1]["mark_samples"] = [c1, c2]
                self.segments[nl1]["dx"] = [1000.0]
                self.segments[nl1]["dy"] = [abs(y[-1] - y[0]) / (len(y) - 1)]
                self.segments[nl1]["d"] = [self.segments[nl1]["dy"]]
                self.segments[nl1]["median1"] = np.median(v)
                self.segments[nl1]["median2"] = 0.
                self.segments[nl1]["x"] = np.round(np.median(x), -3)
                self.segments[nl1]["y"] = np.median(y)
                self.segments[nl1]["mask"] = True
                self.segments[nl1]["block"] = self.n_block
                self.segments[nl1]["dir"] = "odd"
                self.segments[nl1]["pos"] = self.segments[nl1]["x"]
                self.segments[nl1]["direction"] = 0.
                self.segments[nl1]["sensor"] = 1
                x = []
                y = []
                t = []
                height = []
                v = []
            else:
                val = line.split()
                x.append(np.round(float(val[0]), -3))
                y.append(float(val[1]))
                v.append(float(val[14]))
                t.append(float(val[6]))
                height.append(float(val[4]))
        self.sensor1.extend(v)
        self.x.extend(x)
        self.y.extend(y)
        self.z.extend(height)
        self.topo.extend(t)
        self.segments[nl] = {}
        c1 = c2
        c2 += len(x)
        self.segments[nl]["mark_samples"] = [c1, c2]
        self.segments[nl]["dx"] = [1000.0]
        self.segments[nl]["dy"] = [abs(y[-1] - y[0]) / (len(y) - 1)]
        self.segments[nl]["d"] = [self.segments[nl]["dx"]]
        self.segments[nl]["median1"] = np.median(v)
        self.segments[nl]["median2"] = 0.
        self.segments[nl]["x"] = np.round(np.median(x), 3)
        self.segments[nl]["y"] = np.round(np.median(y), 3)
        self.segments[nl]["mask"] = True
        self.segments[nl]["block"] = self.n_block
        self.segments[nl]["dir"] = "odd"
        self.segments[nl]["pos"] = self.segments[nl]["x"]
        self.segments[nl]["direction"] = 0.
        self.segments[nl]["sensor"] = 1
        self.n_lines = nl
        self.x = np.array(self.x)
        self.y = np.array(self.y)
        self.z = np.array(self.z)
        self.topo = np.zeros_like(self.x)
        self.sensor1 = np.array(self.sensor1)
        # Store original data to arrays xxx_ori
        self.sensor1_ori = np.copy(self.sensor1)
        self.time = np.arange(len(self.x))
        self.data = self.lines()
        self.data["grad_data"] = False
        self.data["year"] = 0
        self.data["dispo"] = 0
        self.data["block"] = self.n_block
        self.data["height"] = np.mean(self.z)
        self.data["line_declination"] = 0.0
        return self.data

    def get_line(self, i_line=0):
        """
        Get data of one single line

        Parameters
        ----------
        i_line : int, optional
            Number of line to be extracted (counting starts at 0). The default is 0.

        Returns
        -------
        sensor1: numpy float array
            Data of sensor 1.
        sensor2: numpy float array
            Data of sensor 2.
        x: numpy float array
            X-coordinate of all data points extracted.
        y: numpy float array
            Y-coordinate of all data points extracted.
        z: numpy float array
            height of lower sensor above ground of all data points.
        time: numpy int array
            second of day of acquisition of all data points extracted.
        mask: bool
            mask whether data should be plotted (True) or not (False)
        direction: str
            Line direction with respect to magnetic North (positive from N to E)
        sensor: int
            if 0, two sensors are used in vertical configuration. 1 if only one
            sensor was used or sensor 1 in horizontal configuration. 2: sensor 2
            in horizontal configuration.
        median1: float
            Median value of line for data of sensor1.
        median2: float
            Median value of line for data of sensor2.
        block: int
            Number of data set having been read

        """
        n1 = self.segments[i_line]["mark_samples"][0]
        n2 = self.segments[i_line]["mark_samples"][-1]
        if self.grad_data:
            return (
                self.sensor1[n1:n2],
                self.sensor2[n1:n2],
                self.x[n1:n2],
                self.y[n1:n2],
                self.z[n1:n2],
                self.time[n1:n2],
                self.segments[i_line]["mask"],
                self.segments[i_line]["direction"],
                self.segments[i_line]["sensor"],
                self.segments[i_line]["median1"],
                self.segments[i_line]["median2"],
                self.segments[i_line]["block"],
            )
        dum = np.array([0])
        return (
            self.sensor1[n1:n2],
            dum,
            self.x[n1:n2],
            self.y[n1:n2],
            self.z[n1:n2],
            self.time[n1:n2],
            self.segments[i_line]["mask"],
            self.segments[i_line]["direction"],
            self.segments[i_line]["sensor"],
            self.segments[i_line]["median1"],
            self.segments[i_line]["median2"],
            self.segments[i_line]["block"],
        )

    def lines(self):
        """
        Put all data into a simplified dictionary, one entry per line

        No input parameters

        Returns
        -------
        data: dictionary with one entry per line, key = number of line.
              Each entry is itself a dictionary containing the following entries:

              - "s1": Numpy float array with data of sensor 1
              - "s2": Numpy float array with data of sensor 2
                If only data of one single sensor were measured, "s2" contains
                only one zero.
              - "x":  Numpy float array with E-W coordinates of data points
              - "y":  Numpy float array with N-S coordinates of data points
              - "grad_flag" bool. True if 2 sensors were used, False if only one sensor
              - "mask": bool, True if line should be plotted, False if excluded from plotting
              - "direction": str, direction of a line with respect to magnetic N
                positive from N to E
              - "sensor":
                if 0, two sensors are used in vertical configuration. 1 if only one
                sensor was used or sensor 1 in horizontal configuration. 2: sensor 2
                in horizontal configuration.
              - "median1": float, median of data from sensor 1
              - "median2": float, median of data from sensor 2
              - "block": int, number of data set having been read
        """
        data = {}
        for i, _ in self.segments.items():
            s1, s2, x, y, z, t, m, d, s, med1, med2, block = self.get_line(i)
            data[i] = {}
            data[i]["s1"] = s1
            data[i]["s2"] = s2
            data[i]["x"] = x
            data[i]["y"] = y
            data[i]["z"] = z
            data[i]["time"] = t
            data[i]["mask"] = m
            data[i]["direction"] = d
            data[i]["sensor"] = s
            data[i]["median1"] = med1
            data[i]["median2"] = med2
            data[i]["block"] = block
        return data

    def write_dat(self, file, data1, x, y, *, data2=None, z=None):
        """
        Writes Geometrics magnetic gradiometer file in Surfer format.

        Parameters
        ----------
        Obligatory parameters:

        file : str
            File where to store the data.
        data1 : numpy 1D float array
            Contains all data of sensor 1
        x :     numpy 1D float array
            Contains E-W coordinates of all data points
        y :     numpy 1D float array
            Contains N-S coordinates of all data points

        Optional parameters:

        data2 : numpy 1D float array. Optional, default: None
            Contains all data of sensor 2 if there are (must have the same shape as data1).
        z :     numpy 1D float array. Optional, default. None
            Topography of all data points. If it is None, z-coordinate is set to zero

        Returns
        -------
        None.

        """
        with open(file, "w", encoding="utf-8") as fo:
            fo.write(
                "           X           Y           Z     TOP_RDG  "
                + "BOTTOM_RDG    VRT_GRAD\n"
            )
            # Write gridded data
            if isinstance(data1, np.ndarray):
                if not isinstance(z, np.ndarray):
                    z = np.zeros_like(data1)
                ny, nx = data1.shape
                if not isinstance(data2, np.ndarray):
                    for ix in range(nx):
                        xx = x[ix]
                        for iy in range(ny):
                            fo.write(
                                f"{xx:12.3f}{y[iy]:12.3f}"
                                + f"{z[iy,ix]:12.3f}{data1[iy,ix]:12.3f}\n"
                            )
                else:
                    for ix in range(nx):
                        xx = x[ix]
                        for iy in range(ny):
                            fo.write(
                                f"{xx:12.3f}{y[iy]:12.3f}"
                                + f"{z[iy,ix]:12.3f}{data1[iy,ix]:12.3f}"
                                + f"{data2[iy,ix]:12.3f}"
                                + f"{data2[iy,ix]-data1[iy,ix]:12.3f}\n"
                            )
            # Write non-gridded data
            else:
                for key, val in data1.items():
                    if isinstance(key, str):
                        break
                    if data1["grad_data"]:
                        for i, xx in enumerate(val["x"]):
                            fo.write(
                                f"{xx:12.3f}{val['y'][i]:12.3f}"
                                + f"{val['z'][i]:12.3f}{val['s1'][i]:12.3f}"
                                + f"{val['s2'][i]:12.3f}{val['s2'][i]-val['s1'][i]:12.3f}\n"
                            )
                    else:
                        for i, xx in enumerate(val["x"]):
                            fo.write(
                                f"{xx:12.3f}{val['y'][i]:12.3f}"
                                + f"{val['z'][i]:12.3f}{val['s1'][i]:12.3f}\n"
                            )
        print(f"\nfile {file} written")

    def read_base(self, file, year):
        """
        Wrapper to read base station data

        Parameters
        ----------
        file : str
            Name of data file containing base station data.
        year : int
            Year of data acquisition (often, data contain only day, not year).

        Returns
        -------
        None.

        """
        self.base = Geometrics()
        if not "temp" in file:
            print(f"\nRead base station file {file}")
        self.base.read_base(file, year=year)

    def write_base(self, file):
        """
        Wrapper to write base station data. May be used if time variations are
        calculated from data or if original base station data were modified,
        normally by muting erroneous data

        Parameters
        ----------
        file : str
            Name of output file.

        Returns
        -------
        None.

        """
        self.base.write_base(file)



# -*- coding: utf-8 -*-
"""
Last modified on Sep 04, 2024

@author: Hermann Zeyen <hermann.zeyen@universite-paris-saclay.fr>
         Université Paris-Saclay, France

Contains the following classes for plotting purposes of program PyMaGra:

    - plot
    - newWindow

"""

import os
import numpy as np
from PyQt5.uic import loadUiType
from PyQt5 import QtWidgets
from PyQt5.QtWidgets import QVBoxLayout, QWidget
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg as FigureCanvas,
    NavigationToolbar2QT as NavigationToolbar,
)
from matplotlib.figure import Figure
from matplotlib.ticker import AutoMinorLocator
from matplotlib.path import Path as P
from matplotlib import tri

Ui_MainWindow, QMainWindow = loadUiType(
    os.path.join(os.path.dirname(os.path.abspath(__file__)), "Magnetics.ui")
)


class plot(QMainWindow, Ui_MainWindow):
    """
    Class for plotting in PyMaGra

    Depends on the following packages:

        - numpy
        - PyQt5
        - matplotlib

    Needed private files:

        - Magnetics.ui (user interface, produced with QtDesigner)

    Contains the following methods:

        - __init__
        - setClipPath
        - set_geography_flag
        - toggle_geography_flag
        - plot_geography
        - plot_lineaments
        - plot_north
        - plot_triang
        - plot_image
        - addMPL
        - rmMPL
        - plotFloating

    """

    def __init__(self, main):
        super().__init__()
        self.setupUi(
            self
        )  # Set up main window based on file Magnetics.ui created with QT Designer
        self.fig = Figure()  # create a first figure in central widget
        self.addMPL(self.fig)
        self.main = main
        self.grid_flag = True
        self.plotLin_flag = False
        self.geography_flag = False
        self.legend = []
        self.grad_data = []
        self.unit = "nT"
        self.bar_or = "vertical"
        self.nticks = 10
        self.axfont = 20

    def setClipPath(self, x, y):
        """
        define clip path

        Parameters
        ----------
        x : numpy float array
            contains all x-coordinates of data points.
        y : numpy float array
            contains all y-coordinates of data points.

        Returns
        -------
        PATH matplotlib.path

        """
        clip_x = []
        clip_y = []
        if self.main.direction == 0:
            for xx in np.unique(x):
                yy = np.round(np.nanmax(y[x == xx]), 2)
                if len(clip_x) == 0:
                    yy0 = yy
                else:
                    if yy != yy0:
                        clip_x.append(xx - 0.01)
                        clip_y.append(yy0)
                clip_x.append(xx)
                clip_y.append(yy)
                yy0 = yy
            for xx in np.unique(x)[::-1]:
                yy = np.round(np.nanmin(y[x == xx]), 2)
                if len(clip_x) == 0:
                    yy0 = yy
                else:
                    if yy != yy0:
                        clip_x.append(xx + 0.01)
                        clip_y.append(yy0)
                clip_x.append(xx)
                clip_y.append(yy)
                yy0 = yy
        else:
            for yy in np.unique(y):
                xx = np.round(np.nanmax(x[y == yy]), 2)
                if len(clip_x) == 0:
                    xx0 = xx
                else:
                    if xx != xx0:
                        clip_y.append(yy - 0.01)
                        clip_x.append(xx0)
                clip_y.append(yy)
                clip_x.append(xx)
                xx0 = xx
            for yy in np.unique(y)[::-1]:
                xx = np.round(np.nanmin(x[y == yy]), 2)
                if len(clip_x) == 0:
                    xx0 = xx
                else:
                    if xx != xx0:
                        clip_y.append(yy + 0.01)
                        clip_x.append(xx0)
                clip_y.append(yy)
                clip_x.append(xx)
                xx0 = xx
        clip_x.append(clip_x[0])
        clip_y.append(clip_y[0])
        clip = np.zeros((len(clip_x), 2))
        clip[:, 0] = np.array(clip_x)
        clip[:, 1] = np.array(clip_y)
        # clip contains the coordinates of the clipping path.
        # codes will contain the way to connect clipping points (move to the first point
        #    of the path, draw lines to all other points and finally close the path)
        codes = []
        Path = P
        codes += [Path.MOVETO]
        codes += [Path.LINETO] * (len(clip_x) - 2)
        codes += [Path.CLOSEPOLY]
        # set clipping
        return Path(clip, codes)

    def set_geography_flag(self, option):
        """
        Set flag to plot geography information or not to value "option".

        Parameters
        ----------
        option : bool
            If True, geography information is plotted.

        """
        self.geography_flag = option

    def toggle_geography_flag(self):
        """
        Inverse value of flag to plot geography information or not.

        No input or output parameters

        """
        self.geography_flag = not self.geography_flag

    def plot_geography(self, ax):
        """
        Plot geography information into axis "ax".

        Parameters
        ----------
        ax : Matplotlib.Axis object

        Geography information is stored in class Main as dictionary with entries

        - "type" : may be "point" or "line"
        - "x" : x coordinate(s) of point or line
        - "y" : y coordinate(s) of point or line

        """
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        for key in self.main.geography.keys():
            # Plot a point
            if self.main.geography[key]["type"] == "POINT":
                x = self.main.geography[key]["x"]
                y = self.main.geography[key]["y"]
                # If fpoint outside map don't plot
                if x < xmin or x > xmax or y < ymin or y > ymax:
                    continue
                # Check whether a point has already been plotted. If not, add label to legend
                if "town" in self.legend:
                    ax.plot(x, y, "o", color="black")
                else:
                    self.legend.append("town")
                    ax.plot(x, y, "o", color="black", label="town/place")
                ax.text(
                    x,
                    y,
                    "  " + self.main.geography[key]["name"],
                    ha="left",
                    va="center",
                )
            else:
                # Plot a line and check whether a line has already been plotted.
                if "geology" in self.legend:
                    ax.plot(
                        self.main.geography[key]["x"],
                        self.main.geography[key]["y"],
                        "k",
                    )
                else:
                    self.legend.append("geology")
                    ax.plot(
                        self.main.geography[key]["x"],
                        self.main.geography[key]["y"],
                        "k",
                        label="geology",
                    )

    def plot_lineaments(self, ax):
        """
        Plot lineaments picked from tilt maps into axis "ax".

        Parameters
        ----------
        ax : Matplotlib.Axis object

        Lineament information is stored as dictionary with entries

        - "type" : may be "gravity" or "magnetic"
        - "x" : x coordinate(s) of point or line
        - "y" : y coordinate(s) of point or line

        """
        col = "k"
        for key in self.lineaments.keys():
            if "gravity" in self.lineaments[key]["type"]:
                col = "w"
            elif "magnetic" in self.lineaments[key]["type"]:
                col = "k"
            if self.lineaments[key]["type"] in self.legend:
                ax.plot(
                    self.lineaments[key]["x"],
                    self.lineaments[key]["y"],
                    col,
                    ls="--",
                    linewidth=4,
                )
            else:
                self.legend.append(self.lineaments[key]["type"])
                ax.plot(
                    self.lineaments[key]["x"],
                    self.lineaments[key]["y"],
                    col,
                    label=self.lineaments[key]["type"],
                    ls="--",
                    linewidth=4,
                )

    def plot_north(self, ax, angle, pos):
        """
        Plot North arrow into map
        It is supposed that the map is drawn with set_aspect="equal"

        Parameters
        ----------
        ax : matplotlib axes object
            axis where to plot the North arrow.
        angle : float
            Angle in degrees from N to E (geological sense, not mathematical).
        pos : str
            string of two characters indicationg the position of the arrow.
            One character may be "t" or "b" for top or bottom of the plot
            The other letter may be "l" or "r" for left or right
            The order of the two characters as well as capitalization has no
            importance.

        Returns
        -------
        None.

        """
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        dx = (xmax - xmin) * 0.02
        dy = (ymax - ymin) * 0.05
        alpha = 90.0 - angle
        if alpha < -180:
            alpha += 360.0
        if alpha > 180:
            alpha -= 360.0
        c = np.cos(np.deg2rad(alpha))
        s = np.sin(np.deg2rad(alpha))
        dxa = dy * c
        dya = dy * s
        radius = np.sqrt(dxa**2 + dya**2)

        if "t" in pos.lower():
            y0 = ymax - (1.5 + s) * dy
        else:
            y0 = ymin + (1.5 - s) * dy
        if "r" in pos.lower():
            x0 = xmax - (2.5 + 2 * c) * dx
        else:
            x0 = xmin + (2.5 - 2 * c) * dx

        ax.arrow(
            x0,
            y0,
            dxa,
            dya,
            width=dya / 5.0,
            length_includes_head=True,
            head_length=radius,
            head_width=radius / 1.5,
            shape="left",
            color="k",
            edgecolor=None,
        )
        ax.arrow(
            x0,
            y0,
            dxa,
            dya,
            width=dya / 5.0,
            length_includes_head=True,
            head_length=radius,
            head_width=radius / 1.5,
            shape="right",
            facecolor="w",
            edgecolor="k",
        )
        ax.text(
            x0 + dxa * 1.6,
            y0 + dya * 1.75,
            "N",
            horizontalalignment="center",
            verticalalignment="center_baseline",
            rotation=alpha - 90.0,
            rotation_mode="anchor",
        )

    def plot_sensor_triang(
        self,
        ax,
        x,
        y,
        s,
        vmin,
        vmax,
        percent,
        cmap,
        xlabel,
        ylabel,
        title,
        cbar_title,
        dec=0.0,
    ):
        """
        Plot map of one sensor by triangulation

        Parameters
        ----------
        ax : Matplotlib axis
            Axis where to plot the map
        x : Numpy 1D float array
            X-coordinates of data
        y : Numpy 1D float array
            Y-coordinates of data
        s : Numpy 1D float array
            Values [nT or mGal]
        vmin : float
            Value for lower end of color scale [nT or mGal]
            If vmin==vmax, the color scale is calculated automatically based on
            quantiles (see "percent")
        vmax : float
            Value for upper end of color scale [nT or mGal]
        percent : float
            Scaling of color scale. Maximum value is percentile (1-percent),
            minimum value for percentile percent. This is only used if mincol==maxcol.
        cmap : Matplotlib color map
            Color map to be used
        xlabel : str
            Label for x-axis
        ylabel : str
            Label for y-axis
        title : str
            Title of map
        cbar_title : str
            Text for color bar
        dec : float (optional, default : 0.)
            Direction of the measurement lines with respect to magnetic North
            in degrees, measured from N to E

        Returns
        -------
        None.

        """
        self.legend = []
        # Define clip paths for both sensors, searching for every line maximum and
        #    minimum coordinates
        xmin = x.min()
        xmax = x.max()
        ymin = y.min()
        ymax = y.max()
        clip_path = self.setClipPath(x, y)
        rd = (
            int(np.ceil(-np.log10(np.nanquantile(s, 0.999) - np.nanquantile(s, 0.001))))
            + 2
        )
        # Define collor bar limits and levels
        if vmin == vmax:
            if percent > 0:
                max_col = np.round(np.nanquantile(s, 1 - percent), rd)
                min_col = np.round(np.nanquantile(s, percent), rd)
            else:
                max_col = np.round(np.nanmax(s), rd)
                min_col = np.round(np.nanmin(s), rd)
        else:
            max_col = np.round(vmax, rd)
            min_col = np.round(vmin, rd)
        rd = int(np.ceil(-np.log10(max_col - min_col))) + 2
        nlev = 51
        ds = np.round((max_col - min_col) / nlev, rd)
        levels = np.arange(min_col, max_col + ds / 2, ds)
        # Calculate triangulation
        triang = tri.Triangulation(x, y)
        # Do triangulation plot
        gci = ax.tricontourf(
            triang,
            s,
            extend="both",
            vmin=min_col,
            vmax=max_col,
            levels=levels,
            cmap=cmap,
        )
        ax.set_clip_on(True)
        for collection in gci.collections:
            collection.set_clip_path(clip_path, transform=ax.transData)
        ax.set_title(title)
        ax.set_xlabel(xlabel, fontsize=self.axfont)
        ax.set_ylabel(ylabel, fontsize=self.axfont)
        ax.tick_params(labelsize=self.axfont)
        if self.grid_flag:
            ax.grid(visible=True, which="both")
        if self.plotLin_flag:
            self.plot_lineaments(ax)
        ax.set_xlim([xmin, xmax])
        ax.set_ylim([ymin, ymax])
        ax.xaxis.set_minor_locator(AutoMinorLocator())
        ax.yaxis.set_minor_locator(AutoMinorLocator())
        if self.bar_or == "vertical":
            cax = ax.inset_axes([1.05, 0.05, 0.02, 0.9], transform=ax.transAxes)
        else:
            cax = ax.inset_axes([0.05, -0.15, 0.9, 0.05], transform=ax.transAxes)
        smin = np.round(np.nanmin(s), rd)
        smax = np.round(np.nanmax(s), rd)
        ssmin = levels[0]
        ssmax = levels[-1]
        ds = (ssmax - ssmin) / self.nticks
        ticks = np.round(np.arange(ssmin, ssmax + ds / 2, ds), rd)
        ticks = list(ticks)
        # Plot color bar
        cbar = plt.colorbar(
            gci,
            orientation=self.bar_or,
            cax=cax,
            fraction=0.1,
            ticks=ticks,
            extend="both",
        )
        if self.bar_or == "vertical":
            cbar.ax.set_ylabel(cbar_title)
            cbar.ax.text(
                0.0,
                -0.06,
                f"{smin}",
                verticalalignment="top",
                horizontalalignment="left",
                transform=cbar.ax.transAxes,
            )
            cbar.ax.text(
                0.0,
                1.06,
                f"{smax}",
                verticalalignment="bottom",
                horizontalalignment="left",
                transform=cbar.ax.transAxes,
            )
        else:
            cbar.ax.set_xlabel(cbar_title)
            cbar.ax.text(
                0.0,
                1.0,
                f"{smin}",
                verticalalignment="bottom",
                horizontalalignment="right",
                transform=cbar.ax.transAxes,
            )
            cbar.ax.text(
                1.0,
                1.0,
                f"{smax}",
                verticalalignment="bottom",
                horizontalalignment="left",
                transform=cbar.ax.transAxes,
            )
        ax.set_aspect("equal", adjustable="box")
        if self.geography_flag:
            self.plot_geography(ax)
        if self.plotLin_flag:
            ax.legend(bbox_to_anchor=(1, 1), loc="upper right")
        self.plot_north(ax, -dec, "tl")

    def plot_triang(
        self,
        data,
        title="Measured magnetic data",
        xlabel="Easting",
        ylabel="Northing",
        percent=0.01,
        c="rainbow",
        mincol1=0.0,
        maxcol1=0.0,
        mincol2=0.0,
        maxcol2=0.0,
        mincolg=0.0,
        maxcolg=0.0,
        grad_flag=False,
    ):
        """
        Plot maps of not interpolated data (stn format) of both sensors using
        tricontourf

        Parameters
        ----------
        data: dictionary with one entry per line, key = number of line.
              Each entry is itself a dictionary containing the following entries:

              - "s1": Numpy float array with data of sensor 1
              - "s2": Numpy float array with data of sensor 2
                If only data of one single sensor were measured, "s2" contains only one zero.
              - "x":  Numpy float array with E-W coordinates of data points
              - "y":  Numpy float array with N-S coordinates of data points
              - "grad_flag" bool. True if 2 sensors were used, False if only one sensor
              - "mask": bool, True if line should be plotted, False if excluded from plotting
        title : str, optional
            Plot title for both sensor polts (the text "sensor n" is automatically
            added, n being sensor numnber). The default is "Measured magnetic data".
        xlabel : str, optional
            Text for x-axis. The default is "Easting".
        ylabel : str, optional
            Text for y-axis. The default is "Northing".
        percent : float, optional
            Scaling of color scale. Maximum value is percentile (1-percent),
            minimum value for percentile percent. The default is 0.01.
            This is only used if mincol==maxcol.
        mincol1 : float, optional
            value for lower end of color scale for sensor 1 [nT]
        maxcol1 : float, optional
            value for upper end of color scale for sensor 1 [nT]
            If mincol1==maxcol1, the color scale is calculated automatically based on
            quantiles (see "percent")
        mincol2 : float, optional
            value for lower end of color scale for sensor 2 [nT]
        maxcol2 : float, optional
            value for upper end of color scale for sensor 2 [nT]
            If mincol2==maxcol2, the color scale is calculated automatically based on
            quantiles (see "percent")
        mincolg : float, optional
            value for lower end of color scale for vertical gradient [nT/m]
        maxcolg : float, optional
            value for upper end of color scale for vertical gradient [nT/m]
            If mincolg==maxcolg, the color scale is calculated automatically based on
            quantiles (see "percent")
        c : str, optional
            Color map to be used. The default is "rainbow".
        grad_flag: bool, optional, default: False
            if True, 3 plots are shown on the screen: Sensor1, sensor2 and gradient.
            if False only the two sensors are plotted

        Returns
        -------
        fig : matplotlib.figure
            Figure used for plot.
        ax : matplotlib.axis
            List of axes used for plots. ax has two entrances, one for each sensor

        """
        x = []
        y = []
        v1 = []
        v2 = []
        self.grad_data = data["grad_data"]
        if data["type"] == "magnetic":
            self.unit = "nT"
        else:
            self.unit = "mGal"
        for key, val in data.items():
            try:
                _ = int(key)
                if not val["mask"]:
                    continue
                x += list(val["x"])
                y += list(val["y"])
                v1 += list(val["s1"])
                if self.grad_data:
                    v2 += list(val["s2"])
            except ValueError:
                break
        x = np.array(x)
        y = np.array(y)
        v1 = np.array(v1)
        v2 = np.array(v2)
        ddx = x.max() - x.min()
        ddy = y.max() - y.min()

        if self.grad_data:
            if grad_flag:
                facx = 10 / (3 * ddx)
                facy = 8 / (3 * ddy)
                if facx < facy:
                    fig, ax = plt.subplots(3, 1, figsize=(10, 8), layout="constrained")
                    self.bar_or = "vertical"
                    self.nticks = 10
                else:
                    fig, ax = plt.subplots(1, 3, figsize=(20, 8), layout="constrained")
                    self.bar_or = "horizontal"
                    self.nticks = 6
            else:
                facx = 10 / (3 * ddx)
                facy = 8 / (3 * ddy)
                if facx < facy:
                    fig, ax = plt.subplots(2, 1, figsize=(10, 8), layout="constrained")
                    self.bar_or = "vertical"
                    self.nticks = 10
                else:
                    fig, ax = plt.subplots(1, 2, figsize=(10, 8), layout="constrained")
                    self.bar_or = "horizontal"
                    self.nticks = 6
        else:
            fig, axis = plt.subplots(1, 1, figsize=(10, 8), layout="constrained")
            ax = [axis]
            self.bar_or = "vertical"
            self.nticks = 10
        matplotlib.rcParams.update({"font.size": 20})
        self.axfont = 20
        plt.rc("xtick", labelsize=self.axfont)
        plt.rc("ytick", labelsize=self.axfont)

        # Eliminate possible nans (coming from function clean_data) for both sensors
        x1 = np.copy(x)
        y1 = np.copy(y)
        s1 = np.copy(v1)
        if self.grad_data:
            x2 = np.copy(x)
            y2 = np.copy(y)
            s2 = np.copy(v2)
            x3 = np.copy(x)
            y3 = np.copy(y)
            s3 = s2 - s1

            index = np.isfinite(s1) & np.isfinite(s2)
            x3 = x3[index]
            y3 = y3[index]
            s3 = s3[index]
            x2 = np.delete(x2, np.isnan(s2))
            y2 = np.delete(y2, np.isnan(s2))
            s2 = np.delete(s2, np.isnan(s2))
        x1 = np.delete(x1, np.isnan(s1))
        y1 = np.delete(y1, np.isnan(s1))
        s1 = np.delete(s1, np.isnan(s1))

        # Do plot for sensor 1
        if self.grad_data:
            if ";" in title[:-1]:
                t = title.split(";")
                txt = t[0] + " Sensor 1;" + t[1]
            else:
                txt = title + " Sensor 1"
        else:
            txt = title
        self.plot_sensor_triang(
            ax[0],
            x1,
            y1,
            s1,
            mincol1,
            maxcol1,
            percent,
            c,
            xlabel,
            ylabel,
            txt,
            f"Field strength [{self.unit}]",
            dec=data["line_declination"],
        )
        if self.grad_data:
            # Do plot for sensor 2
            if ";" in title[:-1]:
                t = title.split(";")
                txt = t[0] + " Sensor 2;" + t[1]
            else:
                txt = title + " Sensor 2"
            self.plot_sensor_triang(
                ax[1],
                x2,
                y2,
                s2,
                mincol2,
                maxcol2,
                percent,
                c,
                xlabel,
                ylabel,
                txt,
                f"Field strength [{self.unit}]",
                dec=data["line_declination"],
            )
            # plot vertical gradient
            if grad_flag:
                if ";" in title[:-1]:
                    t = title.split(";")
                    txt = t[0] + " Gradient;" + t[1]
                else:
                    txt = title + " Gradient"
                self.plot_sensor_triang(
                    ax[2],
                    x3,
                    y3,
                    s3,
                    mincolg,
                    maxcolg,
                    percent,
                    c,
                    xlabel,
                    ylabel,
                    txt,
                    f"Field gradient [{self.unit}/m]",
                    dec=data["line_declination"],
                )
        # Erase actual plot in central widget and plot the new one
        self.rmMPL()
        self.addMPL(fig)
        return fig, ax

    def plot_sensor_image(
        self,
        ax,
        x,
        y,
        s,
        vmin,
        vmax,
        percent,
        cmap,
        xlabel,
        ylabel,
        title,
        cbar_title,
        dec=0.0,
    ):
        """
        Plot map of one sensor by triangulation

        Parameters
        ----------
        ax : Matplotlib axis
            Axis where to plot the map
        x : Numpy 1D float array
            X-coordinates of data columns
        y : Numpy 1D float array
            Y-coordinates of data rows
        s : Numpy 2D float array
            Values [nT or mGal] on a regular grid
        vmin : float
            Value for lower end of color scale [nT or mGal]
            If vmin==vmax, the color scale is calculated automatically based on
            quantiles (see "percent")
        vmax : float
            Value for upper end of color scale [nT or mGal]
        percent : float
            Scaling of color scale. Maximum value is percentile (1-percent),
            minimum value for percentile percent. This is only used if mincol==maxcol.
        cmap : Matplotlib color map
            Color map to be used
        xlabel : str
            Label for x-axis
        ylabel : str
            Label for y-axis
        title : str
            Title of map
        cbar_title : str
            Text for color bar
        dec : float (optional, default : 0.)
            Direction of the measurement lines with respect to magnetic North
            in degrees, measured from N to E

        Returns
        -------
        None.

        """
        matplotlib.rcParams.update({"font.size": 20})
        axfont = 20
        matplotlib.rcParams.update({"font.size": 20})
        plt.rc("xtick", labelsize=axfont)
        plt.rc("ytick", labelsize=axfont)
        # Calculate color scale
        if vmin == vmax:
            if percent > 0:
                max_col = np.nanquantile(s, 1 - percent)
                min_col = np.nanquantile(s, percent)
            else:
                max_col = np.nanmax(s)
                min_col = np.nanmin(s)
        else:
            max_col = vmax
            min_col = vmin
        rd = int(np.ceil(-np.log10(max_col - min_col))) + 2
        xmin = np.nanmin(x)
        xmax = np.nanmax(x)
        ymin = np.nanmin(y)
        ymax = np.nanmax(y)
        dx2 = (x[1] - x[0]) / 2
        dy2 = (y[1] - y[0]) / 2
        im1 = ax.imshow(
            np.flip(s, axis=0),
            cmap=cmap,
            aspect="equal",
            extent=[xmin - dx2, xmax + dx2, ymin - dy2, ymax + dy2],
            vmin=min_col,
            vmax=max_col,
        )
        if self.grad_data:
            ax.set_title(title)
        else:
            ax.set_title(title)
        ax.set_xlabel(xlabel, fontsize=axfont)
        ax.set_ylabel(ylabel, fontsize=axfont)
        ax.tick_params(labelsize=axfont)
        if self.grid_flag:
            ax.grid(visible=True, which="both")
        if self.plotLin_flag:
            self.plot_lineaments(ax)
        ax.set_xlim([xmin, xmax])
        ax.set_ylim([ymin, ymax])
        ax.xaxis.set_minor_locator(AutoMinorLocator())
        ax.yaxis.set_minor_locator(AutoMinorLocator())
        if self.bar_or == "vertical":
            cax = ax.inset_axes([1.05, 0.05, 0.02, 0.9], transform=ax.transAxes)
        else:
            cax = ax.inset_axes([0.05, -0.15, 0.9, 0.05], transform=ax.transAxes)
        smin = np.round(np.nanmin(s), rd)
        smax = np.round(np.nanmax(s), rd)
        ssmin = min_col
        ssmax = max_col
        ds = (ssmax - ssmin) / self.nticks
        ticks = np.round(np.arange(ssmin, ssmax + ds / 2, ds), rd)
        ticks = list(ticks)
        labels = []
        for t in ticks:
            labels.append(f"{t}")
        cbar = plt.colorbar(
            im1,
            orientation=self.bar_or,
            cax=cax,
            fraction=0.1,
            ticks=ticks,
            extend="both",
        )
        if self.bar_or == "vertical":
            #            cbar.ax.set_yticks(ticks,labels=labels)
            cbar.ax.set_ylabel(cbar_title)
            cbar.ax.text(
                0.0,
                -0.06,
                f"{smin}",
                verticalalignment="top",
                horizontalalignment="left",
                transform=cbar.ax.transAxes,
            )
            cbar.ax.text(
                0.0,
                1.06,
                f"{smax}",
                verticalalignment="bottom",
                horizontalalignment="left",
                transform=cbar.ax.transAxes,
            )
        else:
            #            cbar.ax.set_xticks(ticks,labels=labels)
            cbar.ax.set_xlabel(cbar_title)
            cbar.ax.text(
                0.0,
                1.0,
                f"{smin}",
                verticalalignment="bottom",
                horizontalalignment="right",
                transform=cbar.ax.transAxes,
            )
            cbar.ax.text(
                1.0,
                1.0,
                f"{smax}",
                verticalalignment="bottom",
                horizontalalignment="left",
                transform=cbar.ax.transAxes,
            )
        if self.geography_flag:
            self.plot_geography(ax)
        #        ax[0].set_aspect('equal', adjustable='box', anchor=anchor)
        ax.set_aspect("equal", adjustable="box")
        if self.plotLin_flag:
            ax.legend(bbox_to_anchor=(1, 1), loc="upper right")
        self.plot_north(ax, -dec, "tl")

    def plot_image(
        self,
        title="Measured magnetic data",
        xlabel="Easting",
        ylabel="Northing",
        percent=0.01,
        c="rainbow",
        mincol1=0,
        maxcol1=0,
        mincol2=0.0,
        maxcol2=0.0,
        mincolg=0.0,
        maxcolg=0.0,
        grad_flag=False,
        dec=0.0,
    ):
        """
        Plot maps of interpolated data of both sensors using imshow

        Parameters
        ----------
        title : str, optional
            Plot title for both sensor polts (the text "sensor n" is automatically
            added, n being sensor numnber). The default is "Measured magnetic data".
        xlabel : str, optional
            Text for x-axis. The default is "Easting".
        ylabel : str, optional
            Text for y-axis. The default is "Northing".
        percent : float, optional
            Scaling of color scale. Maximum value is percentile (1-percent),
            minimum value for percentile percent. The default is 0.01.
        mincol1 : float, optional
            value for lower end of color scale for sensor 1 [nT]
        maxcol1 : float, optional
            value for upper end of color scale for sensor 1 [nT]
            If mincol1==maxcol1, the color scale is calculated automatically based on
            quantiles (see "percent")
        mincol2 : float, optional
            value for lower end of color scale for sensor 2 [nT]
        maxcol2 : float, optional
            value for upper end of color scale for sensor 2 [nT]
            If mincol2==maxcol2, the color scale is calculated automatically based on
            quantiles (see "percent")
        mincolg : float, optional
            value for lower end of color scale for vertical gradient [nT/m]
        maxcolg : float, optional
            value for upper end of color scale for vertical gradient [nT/m]
            If mincolg==maxcolg, the color scale is calculated automatically based on
            quantiles (see "percent")
        c : str, optional
            Color map to be used. The default is "rainbow".
        grad_flag: bool, optional, default: False
            if True, 3 plots are shown on the screen: Sensor1, sensor2 and gradient.
            if False only the two sensors are plotted
        dec : float
            Y direction with respect to geographic north (measured from N to E)

        Returns
        -------
        fig : matplotlib.figure
            Figure used for plot.
        ax : matplotlib.axis
            List of axes used for plots. ax has two entrances, one for each sensor

        """
        self.legend = []
        ddx = self.main.x_inter.max() - self.main.x_inter.min()
        ddy = self.main.y_inter.max() - self.main.y_inter.min()
        if self.grad_data:
            if grad_flag:
                if ddx > 1.5 * ddy:
                    fig, ax = plt.subplots(3, 1, figsize=(10, 8), layout="constrained")
                    self.bar_or = "vertical"
                    #                    anchor = 'E'
                    self.nticks = 10
                else:
                    fig, ax = plt.subplots(1, 3, figsize=(20, 8), layout="constrained")
                    self.bar_or = "horizontal"
                    #                    anchor = 'S'
                    self.nticks = 6
            else:
                if ddx > ddy:
                    fig, ax = plt.subplots(2, 1, figsize=(10, 8), layout="constrained")
                    self.bar_or = "vertical"
                    #                    anchor = 'E'
                    self.nticks = 10
                else:
                    fig, ax = plt.subplots(1, 2, figsize=(10, 8), layout="constrained")
                    self.bar_or = "horizontal"
                    #                    anchor = 'S'
                    self.nticks = 6
        else:
            fig, axis = plt.subplots(1, 1, figsize=(10, 8), layout="constrained")
            self.bar_or = "vertical"
            #            anchor = 'E'
            ax = [axis]
            self.nticks = 10

        # Do plot for sensor 1
        if self.grad_data:
            if ";" in title[:-1]:
                t = title.split(";")
                txt = t[0] + " Sensor 1;" + t[1]
            else:
                txt = title + " Sensor 1"
        else:
            txt = title
        self.plot_sensor_image(
            ax[0],
            self.main.x_inter,
            self.main.y_inter,
            self.main.sensor1_inter,
            mincol1,
            maxcol1,
            percent,
            c,
            xlabel,
            ylabel,
            txt,
            f"Field strength [{self.unit}]",
            dec=dec,
        )

        # Do plot for sensor 2
        if self.grad_data:
            if ";" in title[:-1]:
                t = title.split(";")
                txt = t[0] + " Sensor 2;" + t[1]
            else:
                txt = title + " Sensor 2"
            self.plot_sensor_image(
                ax[1],
                self.main.x_inter,
                self.main.y_inter,
                self.main.sensor2_inter,
                mincol2,
                maxcol2,
                percent,
                c,
                xlabel,
                ylabel,
                txt,
                f"Field strength [{self.unit}]",
                dec=dec,
            )
            # Do plot for gradient
            if grad_flag:
                if ";" in title[:-1]:
                    t = title.split(";")
                    txt = t[0] + " Gradient;" + t[1]
                else:
                    txt = title + " Gradient"
                self.plot_sensor_image(
                    ax[2],
                    self.main.x_inter,
                    self.main.y_inter,
                    self.main.grad_inter,
                    mincolg,
                    maxcolg,
                    percent,
                    c,
                    xlabel,
                    ylabel,
                    txt,
                    f"Gradient [{self.unit}/m]",
                    dec=dec,
                )
        # Erase actual plot in central widget and plot the new one
        self.rmMPL()
        self.addMPL(fig)
        return fig, ax

    def addMPL(self, fig):
        """
        Add widget to actual Figure
        Input: fig is the actually used Figure (matplotlib.figure.Figure object)
        """
        self.canvas = FigureCanvas(fig)
        self.mplvl.addWidget(self.canvas)
        self.canvas.draw()
        self.toolbar = NavigationToolbar(self.canvas, self)
        self.mplvl.addWidget(self.toolbar)

    def rmMPL(
        self,
    ):
        """
        Remove widget from actual Figure
        """
        self.mplvl.removeWidget(self.canvas)
        self.canvas.close()
        self.mplvl.removeWidget(self.toolbar)
        self.toolbar.close()

    def plotFloating(
        self,
        data,
        x,
        y,
        wtitle="",
        sizeh=800,
        sizev=500,
        mincol=0,
        maxcol=0,
        percent=0,
        c="rainbow",
        ptitle="",
        xlabel="",
        ylabel="",
        clabel="",
    ):
        """
        Plot one or several 2D arrays into floating window

        Parameters
        ----------
        data : numpy array
            Data to be plotted.
            The array my be 2D or 3D. If 3D, it is supposed that several figures
            should be plotted into the same window in different frames.
            If more than one plot is to be done, the thrid dimension contains
            the different data sets. Every data set must be 2D. Several data sets
            may be concatenated into data with the following commands:

            - data = data1.reshape(nr,nc,1)
            - data = np.concatenate((data,data2.reshape(nr,nc,1)),axis=2)

            data1 and data2 are arrays with shape (nr,nc) defined on regular grids.

        x : numpy 1D float array
            Positions of the columns
        y : numpy 1D float array
            Positions of the rows
        wtitle : str optional, default: empty string
            Title of floating window
        sizeh : float optional,default: 800
            horizontal size of floating window.
        sizev : float optional, default: 500
            vertical size of floating window
        mincol : float optional, default: 0
            Minimum value of color scale
        maxcol : float optional,  default: 0
            Maximum value of color scale.
            If mincol == maxcol,  and percent == 0, the limits of the color
            scale are minimum and maximum values of the data.
        percent : float optional:
            If >0, mincol and maxcol are calculated as quantiles (percent, 1-percent).
        c: str optional, default: "rainbow"
            color scale to be used.
        ptitle : str, optional, default: empty string
            May be a single string or a list of strings, where the length of the
            list must correspond to the length of the third dimension of data.
        xlabel : str, optional
            Similar as ptitle for lables of horizontal axis
        ylabel : str, optional
            Similar as ptitle for lables of vertical axis
        clabel : str, optional
            Similar as ptitle for lables of color bars

        Returns
        -------
        fig_float,ax_float: Matplot values of figure and axes

        """
        # Only 2D arrays may be plotted, test if 1D data are passed to the routine
        if data.ndim < 2:
            _ = QtWidgets.QMessageBox.warning(
                None,
                "Warning",
                "Function plotFloating is not prepared for 1D plots",
                QtWidgets.QMessageBox.Close,
                QtWidgets.QMessageBox.Close,
            )
            return False
        # Create figure
        fig_float = newWindow(wtitle, sizeh, sizev)
        fig_float.fig.tight_layout()
        # If 2D array is passed create single axis
        if data.ndim == 2:
            ax_float = fig_float.fig.subplots(1, 1)
            data1 = np.copy(data)
            bar_or = "vertical"
            #            anchor = 'E'
            nticks = 10
        # If 3D array is passed create 2 or 3 axis depending on the shape of data
        else:
            ddx = x.max() - x.min()
            ddy = y.max() - y.min()
            data1 = data[:, :, 0]
            if data.shape[2] == 3:
                # If horizontal extension is > 1.5x vertical one, plot axes vertically one
                #    above the next. If not plot axis in horizontal direction
                facx = sizeh / (3 * ddx)
                facy = sizev / (3 * ddy)
                if facx < facy:
                    ax_float = fig_float.fig.subplots(3, 1)
                    bar_or = "vertical"
                    #                    anchor = 'E'
                    nticks = 10
                else:
                    ax_float = fig_float.fig.subplots(1, 3)
                    bar_or = "horizontal"
                    #                    anchor = 'S'
                    nticks = 6
            else:
                facx = sizeh / (2 * ddx)
                facy = sizev / (2 * ddy)
                if facx < facy:
                    ax_float = fig_float.fig.subplots(2, 1)
                    bar_or = "vertical"
                    #                    anchor = 'E'
                    nticks = 10
                else:
                    ax_float = fig_float.fig.subplots(1, 2)
                    bar_or = "horizontal"
                    #                    anchor = 'S'
                    nticks = 6
        # Calculate color scale for both sensors independently
        rd = int(np.ceil(-np.log10(np.nanmax(abs(data1))))) + 2
        if mincol == maxcol:
            if percent > 0:
                max_col = np.round(np.nanquantile(data1, 1 - percent), rd)
                min_col = np.round(np.nanquantile(data1, percent), rd)
            else:
                max_col = np.nanmax(data1)
                min_col = np.nanmin(data1)
        else:
            max_col = maxcol
            min_col = mincol
        # Plot first into axis ax_float[0]
        try:
            ax = ax_float[0]
            pt = ptitle[0]
            xl = xlabel[0]
            yl = ylabel[0]
            cl = clabel[0]
        # If only one subplot has been created
        except TypeError:
            ax = ax_float
            pt = ptitle
            xl = xlabel
            yl = ylabel
            cl = clabel
        dx2 = (x[1] - x[0]) / 2
        dy2 = (y[1] - y[0]) / 2
        im1 = ax.imshow(
            np.flip(data1, axis=0),
            cmap=c,
            aspect="equal",
            extent=[np.min(x) - dx2, np.max(x) + dx2, np.min(y) - dy2, np.max(y) + dy2],
            vmin=min_col,
            vmax=max_col,
        )
        ax.set_title(pt)
        ax.set_xlabel(xl)
        ax.set_ylabel(yl)
        if bar_or == "vertical":
            cax = ax.inset_axes([1.015, 0.05, 0.015, 0.9], transform=ax.transAxes)
        else:
            cax = ax.inset_axes([0.05, -0.15, 0.9, 0.05], transform=ax.transAxes)
        smin = np.round(np.nanmin(data1), rd)
        smax = np.round(np.nanmax(data1), rd)
        ssmin = min_col
        ssmax = max_col
        ds = (ssmax - ssmin) / nticks
        ticks = np.round(np.arange(ssmin, ssmax + ds / 2, ds), rd)
        ticks = list(ticks)
#        cbar = plt.colorbar(
        cbar = fig_float.fig.colorbar(
            im1, orientation=bar_or, ax=ax, cax=cax, fraction=0.1, extend="both", ticks=ticks
        )
        if bar_or == "vertical":
            cbar.ax.set_ylabel(cl)
            cbar.ax.text(
                0.0,
                -0.075,
                f"{smin}",
                verticalalignment="top",
                horizontalalignment="left",
                transform=cbar.ax.transAxes,
            )
            cbar.ax.text(
                0.0,
                1.075,
                f"{smax}",
                verticalalignment="bottom",
                horizontalalignment="left",
                transform=cbar.ax.transAxes,
            )
        else:
            cbar.ax.set_xlabel(cl)
            cbar.ax.text(
                0.0,
                1.0,
                f"{smin}",
                verticalalignment="bottom",
                horizontalalignment="right",
                transform=cbar.ax.transAxes,
            )
            cbar.ax.text(
                1.0,
                1.0,
                f"{smax}",
                verticalalignment="bottom",
                horizontalalignment="left",
                transform=cbar.ax.transAxes,
            )
        if self.grid_flag:
            ax.grid(visible=True)
        #        ax.set_aspect('equal', adjustable='box', anchor=anchor)
        ax.set_aspect("equal", adjustable="box")
        # If more than one map has to be plotted, do this now
        if data.ndim > 2:
            if data.shape[2] > 1:
                data1 = data[:, :, 1]
                rd = int(np.ceil(-np.log10(np.nanmax(abs(data1))))) + 2
                if mincol == maxcol:
                    if percent > 0:
                        max_col = np.round(np.nanquantile(data1, 1 - percent), rd)
                        min_col = np.round(np.nanquantile(data1, percent), rd)
                    else:
                        max_col = np.nanmax(data1)
                        min_col = np.nanmin(data1)
                else:
                    max_col = maxcol
                    min_col = mincol
                im1 = ax_float[1].imshow(
                    np.flip(data1, axis=0),
                    cmap=c,
                    aspect="equal",
                    extent=[
                        np.min(x) - dx2,
                        np.max(x) + dx2,
                        np.min(y) - dy2,
                        np.max(y) + dy2,
                    ],
                    vmin=min_col,
                    vmax=max_col,
                )
                ax_float[1].set_title(ptitle[1])
                ax_float[1].set_xlabel(xlabel[1])
                ax_float[1].set_ylabel(ylabel[1])
                if self.grid_flag:
                    ax_float[1].grid(visible=True)
                if bar_or == "vertical":
                    cax = ax_float[1].inset_axes(
                        [1.015, 0.05, 0.015, 0.9], transform=ax_float[1].transAxes
                    )
                else:
                    cax = ax_float[1].inset_axes(
                        [0.05, -0.15, 0.9, 0.05], transform=ax_float[1].transAxes
                    )
                smin = np.round(np.nanmin(data1), rd)
                smax = np.round(np.nanmax(data1), rd)
                ssmin = min_col
                ssmax = max_col
                ds = (ssmax - ssmin) / nticks
                ticks = np.round(np.arange(ssmin, ssmax + ds / 2, ds), rd)
                ticks = list(ticks)
#                cbar = plt.colorbar(
                cbar = fig_float.fig.colorbar(
                    im1,
                    orientation=bar_or,
                    ax=ax,
                    cax=cax,
                    fraction=0.1,
                    extend="both",
                    ticks=ticks,
                )
                if bar_or == "vertical":
                    cbar.ax.set_ylabel(clabel[1])
                    cbar.ax.text(
                        0.0,
                        -0.075,
                        f"{smin}",
                        verticalalignment="top",
                        horizontalalignment="left",
                        transform=cbar.ax.transAxes,
                    )
                    cbar.ax.text(
                        0.0,
                        1.075,
                        f"{smax}",
                        verticalalignment="bottom",
                        horizontalalignment="left",
                        transform=cbar.ax.transAxes,
                    )
                else:
                    cbar.ax.set_xlabel(clabel[1])
                    cbar.ax.text(
                        0.0,
                        1.0,
                        f"{smin}",
                        verticalalignment="bottom",
                        horizontalalignment="right",
                        transform=cbar.ax.transAxes,
                    )
                    cbar.ax.text(
                        1.0,
                        1.0,
                        f"{smax}",
                        verticalalignment="bottom",
                        horizontalalignment="left",
                        transform=cbar.ax.transAxes,
                    )
                #                ax_float[1].set_aspect('equal', adjustable='box', anchor=anchor)
                ax_float[1].set_aspect("equal", adjustable="box")
                if data.shape[2] > 2:
                    data1 = data[:, :, 2]
                    rd = int(np.ceil(-np.log10(np.nanmax(abs(data1))))) + 2
                    if mincol == maxcol:
                        if percent > 0:
                            max_col = np.round(np.nanquantile(data1, 1 - percent), rd)
                            min_col = np.round(np.nanquantile(data1, percent), rd)
                        else:
                            max_col = np.nanmax(data1)
                            min_col = np.nanmin(data1)
                    else:
                        max_col = maxcol
                        min_col = mincol
                    im1 = ax_float[2].imshow(
                        np.flip(data1, axis=0),
                        cmap=c,
                        aspect="equal",
                        extent=[
                            np.min(x) - dx2,
                            np.max(x) + dx2,
                            np.min(y) - dy2,
                            np.max(y) + dy2,
                        ],
                        vmin=min_col,
                        vmax=max_col,
                    )
                    ax_float[2].set_title(ptitle[2])
                    ax_float[2].set_xlabel(xlabel[2])
                    ax_float[2].set_ylabel(ylabel[2])
                    if self.grid_flag:
                        ax_float[2].grid(visible=True)
                    if bar_or == "vertical":
                        cax = ax_float[2].inset_axes(
                            [1.015, 0.05, 0.015, 0.9], transform=ax_float[2].transAxes
                        )
                    else:
                        cax = ax_float[2].inset_axes(
                            [0.05, -0.15, 0.9, 0.05], transform=ax_float[2].transAxes
                        )
                    smin = np.round(np.nanmin(data1), rd)
                    smax = np.round(np.nanmax(data1), rd)
                    ssmin = min_col
                    ssmax = max_col
                    ds = (ssmax - ssmin) / nticks
                    ticks = np.round(np.arange(ssmin, ssmax + ds / 2, ds), rd)
                    ticks = list(ticks)
#                    cbar = plt.colorbar(
                    cbar = fig_float.fig.colorbar(
                        im1,
                        orientation=bar_or,
                        ax=ax,
                        cax=cax,
                        fraction=0.1,
                        extend="both",
                        ticks=ticks,
                    )
                    if bar_or == "vertical":
                        cbar.ax.set_ylabel(clabel[2])
                        cbar.ax.text(
                            0.0,
                            -0.075,
                            f"{smin}",
                            verticalalignment="top",
                            horizontalalignment="left",
                            transform=cbar.ax.transAxes,
                        )
                        cbar.ax.text(
                            0.0,
                            1.075,
                            f"{smax}",
                            verticalalignment="bottom",
                            horizontalalignment="left",
                            transform=cbar.ax.transAxes,
                        )
                    else:
                        cbar.ax.set_xlabel(clabel[2])
                        cbar.ax.text(
                            0.0,
                            1.0,
                            f"{smin}",
                            verticalalignment="bottom",
                            horizontalalignment="right",
                            transform=cbar.ax.transAxes,
                        )
                        cbar.ax.text(
                            1.0,
                            1.0,
                            f"{smax}",
                            verticalalignment="bottom",
                            horizontalalignment="left",
                            transform=cbar.ax.transAxes,
                        )
                    #       ax_float[2].set_aspect('equal', adjustable='box', anchor=anchor)
                    ax_float[2].set_aspect("equal", adjustable="box")
        fig_float.show()
        return fig_float, ax_float


class newWindow(QWidget):
    """
    This "window" is a QWidget. If it has no parent, it will appear as a
    free-floating window.
    """

    def __init__(self, title, xsize=1800, ysize=1200):
        super().__init__()
        self.fig = Figure()
        self.layout = QVBoxLayout()
        self.setLayout(self.layout)
        self.resize(xsize, ysize)
        self.setWindowTitle(title)
        self.canvas = FigureCanvas(self.fig)
        self.layout.addWidget(self.canvas)
        self.canvas.draw()
        self.toolbar = NavigationToolbar(self.canvas, self)
        self.layout.addWidget(self.toolbar)



# -*- coding: utf-8 -*-
"""
Last modified on Aug 26, 2024

@author: Hermann Zeyen University Paris-Saclay, France
"""

import numpy as np
from PyQt5 import QtWidgets
from ..utilities.utilities import Utilities


class Geometrics:
    """
    Input and output of Geometrics format files.

    Methods:

    - __init__
    - fill_dict
    - read_stn
    - write_stn
    - clean_data
    - line_days
    - get_line
    - geometrics_lines
    - get_segment
    - read_base
    - write_base

    """

    def __init__(self):
        """
        Class Geometrics reads and writes Geometrics .stn files and allows
        a number of data treatments

        Parameters
        ----------

        Returns
        -------
        None.

        """
        # Initialize main frame widget
        #        super(Geometrics, self).__init__()
        super().__init__()
        self.u = Utilities(self)
        self.n_lines = 0
        self.n_data = 0
        self.grad_data = True
        self.geography_flag = False
        self.grid_flag = True
        self.unit = "nT"
        # \        self.read_stn(file)
        self.lines_per_day = {}
        self.grad = np.array([])
        self.year_base = np.array([])
        self.month_base = np.array([])
        self.day_base = np.array([])
        self.hour_base = np.array([])
        self.minute_base = np.array([])
        self.second_base = np.array([])
        self.time_base = np.array([])
        self.jday_base = np.array([])
        self.base = np.array([])
        self.year = np.array([])
        self.month = np.array([])
        self.day = np.array([])
        self.hour = np.array([])
        self.minute = np.array([])
        self.second = np.array([])
        self.time = np.array([])
        self.x = np.array([])
        self.y = np.array([])
        self.x1 = np.array([])
        self.y1 = np.array([])
        self.x2 = np.array([])
        self.y2 = np.array([])
        self.sensor1 = np.array([])
        self.sensor2 = np.array([])
        self.sensor1_ori = np.array([])
        self.sensor2_ori = np.array([])
        self.grad_ori = np.array([])
        self.n_blocks = 0
        self.d_sensor = 0.0
        self.d_sensor2 = 0.0
        self.h_sensor = 0.0
        self.dispo = 0
        self.infile = ""
        self.segments = {}
        self.grad_flag = False
        self.dx = 0.0
        self.dy = 0.0
        self.direction = 0
        self.d_lines = 0.0
        self.line_pos = np.array([])
        self.line_choice = "all"

    def fill_dict(
        self, ntracks, x1, y1, mag1, x2, y2, mag2, mark_samples, dx_seg, dy_seg, d_seg
    ):
        """
        Fill in values of dictionary self.segments for a full line.

        Parameters
        ----------
        ntracks : int
            Actual line number (in horizontal configuration, dictionary entrancies
            ntracks and ntarcks+1 are filled).
        x1 : list of floats
            X-coordinates for sensor 1 data.
        y1 : list of floats
            Y-coordinates for sensor 1 data.
        mag1 : list of floats
            Magnetic data recorded by sensor 1.
        x2 : list of floats
            X-coordinates for sensor 2 data.
        y2 : list of floats
            Y-coordinates for sensor 2 data.
        mag2 : list of floats
            Magnetic data recorded by sensor 1.
        mark_samples : list of int
            Number of the sample just after a mark.
        dx_seg : list of floats
            Step length in x-direction for every segment.
        dy_seg : list of floats
            Step length in y-direction for every segment.
        d_seg : list of floats
            Total step length for every segment.

        Returns
        -------
        None.

        """
        dx = x1[-1] - x1[0]
        dy = y1[-1] - y1[0]
        nt1 = ntracks + 1
        if ntracks == 0:
            nshift = 0
        else:
            nshift = self.segments[ntracks - 1]["mark_samples"][-1]
        if abs(dx) > abs(dy):
            if dx > 0:
                self.segments[ntracks]["direction"] = 90.
            else:
                self.segments[ntracks]["direction"] = -90.
        else:
            if dy > 0:
                self.segments[ntracks]["direction"] = 0.
            else:
                self.segments[ntracks]["direction"] = 180.
        self.segments[ntracks]["median1"] = np.median(mag1)
        self.segments[ntracks]["mask"] = True
        self.segments[ntracks]["block"] = self.n_blocks
        self.segments[ntracks]["mark_samples"] = (
            np.array(mark_samples, dtype=int) + nshift
        )
        self.segments[ntracks]["dx"] = np.array(dx_seg)
        self.segments[ntracks]["dy"] = np.array(dy_seg)
        self.segments[ntracks]["d"] = np.array(d_seg)
        self.segments[ntracks]["x"] = np.median(x1)
        self.segments[ntracks]["y"] = np.median(y1)
        if self.dispo == 0:
            self.segments[ntracks]["median2"] = np.median(mag2)
            self.segments[ntracks]["sensor"] = 0
        else:
            self.segments[ntracks]["sensor"] = 1
            self.segments[ntracks]["median2"] = 0.0
            self.segments[nt1]["median1"] = np.median(mag2)
            self.segments[nt1]["median2"] = 0.0
            self.segments[nt1]["x"] = np.median(x2)
            self.segments[nt1]["y"] = np.median(y2)
            self.segments[nt1]["mask"] = True
            self.segments[nt1]["block"] = self.n_blocks
            self.segments[nt1]["sensor"] = 2
            self.segments[nt1]["direction"] = self.segments[ntracks]["direction"]
            self.segments[nt1]["mark_samples"] = (
                np.array(mark_samples, dtype=int)
                + self.segments[ntracks]["mark_samples"][-1]
            )
            self.segments[nt1]["dx"] = np.array(dx_seg)
            self.segments[nt1]["dy"] = np.array(dy_seg)
            self.segments[nt1]["d"] = np.array(d_seg)

    def append_data(self, x, y, d, year, month, day, hour, minute, second, time):
        """
        Appends date and time information of one segment to global arrays

        Parameters
        ----------
        x: list of float
            x-coordinate of data points
        y: list of float
            y-coordinate of data points
        d: list of float
            measured values of sensor 1
        year: list of int
            year of acquisition of data points
        month: list of int
            month of acquisition of data points
        day: list of int
            day of acquisition of data points
        hour: list of int
            hour of acquisition of data points
        minute: list of int
            minute of acquisition of data points
        second: list of float
            second of acquisition of data points
        time: list of int
            julian day of data points
        """
        self.x = np.concatenate((self.x, np.round(np.array(x, dtype=float), 3)))
        self.y = np.concatenate((self.y, np.round(np.array(y, dtype=float), 3)))
        self.sensor1 = np.concatenate((self.sensor1, np.array(d, dtype=float)))
        self.day = np.concatenate((self.day, np.array(day, dtype=int)))
        self.month = np.concatenate((self.month, np.array(month, dtype=int)))
        self.year = np.concatenate((self.year, np.array(year, dtype=int)))
        self.hour = np.concatenate((self.hour, np.array(hour, dtype=int)))
        self.minute = np.concatenate((self.minute, np.array(minute, dtype=int)))
        self.second = np.concatenate((self.second, np.array(second, dtype=float)))
        tt = (
            np.array(time, dtype=int) * 86400.0
            + np.array(hour, dtype=int) * 3600.0
            + np.array(minute, dtype=int) * 60.0
            + np.array(second, dtype=float)
        )
        self.time = np.concatenate((self.time, tt))

    def read_stn(self, infile, n_block, d_sensor=0.9, h_sensor=0.4, dispo=0):
        """
        Reads a 2-sensor Geometix .stn file
        If other data have already been read from another file, the new ones are
        added to the vectors with new line numbers, even if effectively, the
        measurements may have been done in the prolongation of existing lines.

        Parameters
        ----------
        infile : str
            Name of file to be read.
        n_block : int
            Number of block (file) of this data set within all files to be read.
        d_sensor : float. Optional, default: 0.9 m.
            Distance between sensors [m]. It is supposed that sensor1 is the
            upper one. If this is not the case, d_sensor must be negative.
        h_sensor : float, optional
            Hight of lower sensor (sensor 2) above ground [m]. Default: 0.4 m
        disop : int, optional
            If two sensors were used, dispo indicates there relative position:

            - If dispo == 0: vertical disposition, sensor 2 above sensor 1.
            - If dispo == 1: horizontal disposition, sensor 2 to the right of sensor 1.

        Data are stored in the following arrays:

        - self.sensor1 : 1D numpy float array contains all data of sensor 1
        - self.sensor2 : 1D numpy float array contains all data of sensor 2
        - self.day     : 1D numpy int array contains day of acquisition for all data
        - self.month   : 1D numpy int array contains month of acquisition for all data
        - self.year    : 1D numpy int array contains year of acquisition for all data
        - self.hour    : 1D numpy int array contains hour of acquisition for all data
        - self.minute  : 1D numpy int array contains minute of acquisition for all data
        - self.second  : 1D numpy float array contains second of acquisition for all data
        - self.time    : 1D numpy float array contains seconds of acquisition for all data
          counted from the beginning of the acquisition year
        - self.x       : 1D numpy float array contains x-coordinates of measurement points
        - self.y       : 1D numpy float array contains y-coordinates of measurement points
        - self.segments: Dictionary with information for all lines and segments measured

            Since all data are stored sequentially, we need to know  which data belong
            to which line and it may also be necessary to know where marks were set, thus
            where a measured segment starts and ends.
            self.segments.keys() contains line numbers starting at zero
            Every dictionary entrance contains the following lists:

            - "mark_samples" contains for every mark (starting and
              end of line included) the number of the sample in self.sensor1 or 2
              recorded just after the mark
            - "mask" is initially set to True but may be modified in order to
              plot only certain lines. If True, lines are plotted, else they are
              excluded from plot
            - "median1" : median value of sensor 1 for a line
            - "median2" : median value of sensor 2 for a line
              If sensors are in horizontal configuration, both sensors are treated
              as sensor 1 and "median2" has value 0.
            - "sensor" : may be 0 if both sensors are used in vertical configuration,
              1 or 2 if line has been measured by sensor 1 or 2 in horizontal
              configuration.
            - "dx" contains for every segment the sampling step in X direction.
            - "dy" contains for every segment the sampling step in Y direction.
            - "d" contains for every segment the sampling step length.
              dx, dy and d contain one value less than mark_samples
            - "x" median of x-coordinates
            - "y" median of y-coordiantes
              depending on the line direction, "x" or "y" are supposed to be
              the position of the line
            - "pos" coordinate of the line
            - "block" number of block (or file) if several files are read
            - "dir" may be "odd" or "even". First measured line is "odd"
              (odd or even refers to natural counting)
            - "direction" is "N", "S", "W" or "E" depending on the line direction

            In addition, the following attributes are accessible:

            - self.n_data   : int, total number of data
            - self.n_lines : int, number of lines read
            - self.direction : int, direction of measurement lines.
              May be 0 (Y-direction) or 1 (X-direction)
            - self.d_lines : float, distance between measurement lines

        Returns
        -------
        data: dictionary with one entry per line, key = number of line.
              Each entry is itself a dictionary containing the following entries:

              - "s1": Numpy float array with data of sensor 1
              - "s2": Numpy float array with data of sensor 2
                If only data of one single sensor were measured, "s2" contains
                only one zero.
              - "x":  E-W coordinates of data points
              - "y"   N-S coordinates of data points

        """
        # Read magnetic field data ("stn" file)
        self.n_blocks = n_block
        self.d_sensor = d_sensor
        self.d_sensor2 = d_sensor / 2.0
        self.h_sensor = h_sensor
        self.dispo = dispo
        self.infile = infile
        mag1 = []
        mag2 = []
        xpt = []
        ypt = []
        xpt1 = []
        ypt1 = []
        xpt2 = []
        ypt2 = []
        year_m = []
        month_m = []
        day_m = []
        hour_m = []
        min_m = []
        sec_m = []
        time_m = []
        if self.n_lines == 0:
            self.sensor1 = np.array([])
            self.sensor2 = np.array([])
            self.x = np.array([])
            self.y = np.array([])
            self.x1 = np.array([])
            self.y1 = np.array([])
            self.x2 = np.array([])
            self.y2 = np.array([])
            self.day = np.array([], dtype=int)
            self.month = np.array([], dtype=int)
            self.year = np.array([], dtype=int)
            self.hour = np.array([], dtype=int)
            self.minute = np.array([], dtype=int)
            self.second = np.array([])
            self.time = np.array([])
            self.segments = {}
            n_start = 0
            first = True
        else:
            n_start = len(self.sensor1)
            first = False

        ntracks = self.n_lines - 1

        # read all lines in magnetic data file
        with open(infile, "r", encoding="utf-8") as fh:
            ll = fh.readlines()
        try:
            _ = int(ll[0].split()[0])
        except Exception as e:
            # If none of the above, stop program
            _ = QtWidgets.QMessageBox.critical(
                None,
                "Error",
                f"Error: {e}\n\n"
                + f"File {infile} is not a Geometrics file.\nProgram stops",
                QtWidgets.QMessageBox.Ok,
            )
            raise IOError("Wrong file type\n") from e
        #            sys.exit()
        # store reversed order in list "lines"
        lines = ll[::-1]
        nlines = len(lines)
        last_mark = 5
        ndat_seg = 0
        xx0 = 0.0
        yy0 = 0.0
        ddx = 0.0
        ddy = 0.0
        nstart = True
        nt1 = ntracks
        mark_samples = []
        dx_seg = []
        dy_seg = []
        d_seg = []

        # read the lines and get data values and corresponding coordinates
        for k in range(nlines):
            line = lines[k].split()
            # if first number in line is 3, it is a mark (including beginning and end of line)
            #    seg_break_data contains number of data point before which mark is located
            #    id_seg_break contains the type of mark
            #    seg_break_line contains the line in the data file where mark is located
            #    line segments contains for every line the position of marks (between marks,
            #                  data positions are interpolated at regular spacing)
            #    dx_seg contains distance between samples for every segment
            if int(line[0]) == 3:
                xx = np.round(float(line[1]), 3)
                yy = np.round(float(line[2]), 3)
                # code 36 indicates beginning of a new line
                # code 4 indicates a mark, 5 is the Endline mark
                if int(line[8]) in (4, 5):
                    if ndat_seg == 0:
                        continue
                    mark_samples.append(len(mag1) + n_start)
                    n1 = mark_samples[-2]
                    n2 = mark_samples[-1]
                    ndat_segment = n2 - n1
                    if ndat_segment == 0:
                        del mark_samples[-1]
                        continue
                    # print(f"nline: {ntracks}: nseg: "+\
                    #       f"{len(self.segments[ntracks]['mark_samples'])-1},"+\
                    #       f" ndseg: {ndat_segment}")
                    # If last mark had code 36, a new line was started and it is
                    #    the first mark after start. The number of measurement steps
                    #    is 1 less than the number of data points, if not,
                    #    both are equal. This means that it is supposed that the
                    #    mark was pressed atthe position of the measurement just
                    #    before the mark
                    if nstart:
                        dx = (xx - xx0) / (ndat_segment - 1)
                        dy = (yy - yy0) / (ndat_segment - 1)
                        n2 -= 1
                    else:
                        dx = (xx - xx0) / ndat_segment
                        dy = (yy - yy0) / ndat_segment
                    d = np.sqrt(dx * dx + dy * dy)
                    dx_seg.append(dx)
                    dy_seg.append(dy)
                    d_seg.append(d)
                    if self.dispo > 0:
                        ddx1 = ddx
                        ddy1 = ddy
                        nt1 = ntracks + 1
                        ddx = dy / d * self.d_sensor2
                        ddy = dx / d * self.d_sensor2
                    else:
                        nt1 = ntracks
                        ddx1 = 0.0
                        ddy1 = 0.0
                        ddx = 0.0
                        ddy = 0.0
                    if nstart:
                        xpt1[0] -= ddx
                        ypt1[0] += ddy
                        xpt2[0] += ddx
                        ypt2[0] -= ddy
                    else:
                        xpt1.append(xpt1[-1] + dx - ddx + ddx1)
                        ypt1.append(ypt1[-1] + dy + ddy - ddy1)
                        xpt2.append(xpt2[-1] + dx + ddx - ddx1)
                        ypt2.append(ypt2[-1] + dy - ddy + ddy1)
                        n1 += 1
                    for _ in range(n1, n2):
                        xpt1.append(xpt1[-1] + dx)
                        ypt1.append(ypt1[-1] + dy)
                        xpt2.append(xpt2[-1] + dx)
                        ypt2.append(ypt2[-1] + dy)
                    last_mark = int(line[8])
                    xx0 = xx
                    yy0 = yy
                    nstart = False
                if int(line[8]) == 5 or (int(line[8]) == 36 and last_mark != 5):
                    self.fill_dict(
                        ntracks,
                        xpt1,
                        ypt1,
                        mag1,
                        xpt2,
                        ypt2,
                        mag2,
                        mark_samples,
                        dx_seg,
                        dy_seg,
                        d_seg,
                    )
                    self.append_data(
                        xpt1,
                        ypt1,
                        mag1,
                        year_m,
                        month_m,
                        day_m,
                        hour_m,
                        min_m,
                        sec_m,
                        time_m,
                    )
                    if self.dispo:
                        self.append_data(
                            xpt2,
                            ypt2,
                            mag2,
                            year_m,
                            month_m,
                            day_m,
                            hour_m,
                            min_m,
                            sec_m,
                            time_m,
                        )
                        self.n_lines += 1
                    else:
                        self.sensor2 = np.concatenate(
                            (self.sensor2, np.array(mag2, dtype=float))
                        )
                    last_mark = int(line[8])
                    ndat_seg = 0
                # xpt and ypt contain the coordinates of all data points of the actual line
                if int(line[8]) == 36:
                    xx0 = xx
                    yy0 = yy
                    xpt1 = [xx0]
                    ypt1 = [yy0]
                    xpt2 = [xx0]
                    ypt2 = [yy0]
                    dx_seg = []
                    dy_seg = []
                    d_seg = []
                    mag1 = []
                    mag2 = []
                    time_m = []
                    month_m = []
                    day_m = []
                    hour_m = []
                    min_m = []
                    sec_m = []
                    ndat_segment = int(line[5])
                    ndat_seg = 0
                    nstart = True
                    ntracks = nt1 + 1
                    self.n_lines += 1
                    self.segments[ntracks] = {}
                    mark_samples = [0]
                    if self.dispo:
                        self.segments[ntracks + 1] = {}
                    continue
                # Don't take into account other mark codes (mainly Pause)
                continue
            # if first number in line is 0, it is a data point
            # mag1 contains data of first sensor, mag 2 of second sensor
            # In addition, times are stored
            if int(line[0]) == 0:
                mag1.append(float(line[1]))
                mag2.append(float(line[2]))
                t = line[3].split(":")
                hour_m.append(int(t[0]))
                min_m.append(int(t[1]))
                sec_m.append(float(t[2]))
                t = line[4].split("/")
                day_m.append(int(t[1]))
                month_m.append(int(t[0]))
                year_m.append(int(t[2]))
                time_m.append(self.u.date2julian(day_m[-1], month_m[-1], year_m[-1]))
                ndat_seg += 1

        if self.n_lines > 0 and last_mark != 5:
            self.fill_dict(
                ntracks,
                xpt1,
                ypt1,
                mag1,
                xpt2,
                ypt2,
                mag2,
                mark_samples,
                dx_seg,
                dy_seg,
                d_seg,
            )
            self.append_data(
                xpt1, ypt1, mag1, year_m, month_m, day_m, hour_m, min_m, sec_m, time_m
            )
            if self.dispo > 0:
                self.append_data(
                    xpt2,
                    ypt2,
                    mag2,
                    year_m,
                    month_m,
                    day_m,
                    hour_m,
                    min_m,
                    sec_m,
                    time_m,
                )
            else:
                self.sensor2 = np.concatenate(
                    (self.sensor2, np.array(mag2, dtype=float))
                )
        self.n_data += len(mag1)
        if len(self.sensor1) == len(self.sensor2):
            self.grad = (self.sensor1 - self.sensor2) / self.d_sensor
            self.grad_flag = True
        else:
            self.grad = np.array([0])
            self.grad_flag = False
        del mag1, mag2, xpt, ypt, day_m, month_m, year_m, hour_m, min_m, sec_m
        self.dx = self.segments[0]["dx"][0]
        self.dy = self.segments[0]["dy"][0]
        if first:
            if abs(self.x[0] - self.x[self.segments[0]["mark_samples"][-1]]) > abs(
                self.y[0] - self.y[self.segments[0]["mark_samples"][-1]]
            ):
                self.direction = 1
                self.d_lines = np.round(
                    (self.y.max() - self.y.min()) / (self.n_lines - 1), 1
                )
            else:
                self.direction = 0
                self.d_lines = np.round(
                    (self.x.max() - self.x.min()) / (self.n_lines - 1), 1
                )
        # Store original data to arrays xxx_ori
        self.sensor1_ori = np.copy(self.sensor1)
        self.sensor2_ori = np.copy(self.sensor2)
        self.grad_ori = np.copy(self.grad)
        self.line_pos = np.zeros(len(self.segments.keys()))
        if self.direction == 1:
            for key, entries in self.segments.items():
                self.line_pos[key] = entries["y"]
                self.segments[key]["pos"] = entries["y"]
        else:
            for key, entries in self.segments.items():
                self.line_pos[key] = entries["x"]
                self.segments[key]["pos"] = entries["x"]
        self.line_choice = "all"
        if self.direction == 0:
            sens = np.sign(self.y[1] - self.y[0])
            for key, entries in self.segments.items():
                i1 = entries["mark_samples"][0]
                if np.sign(self.y[i1 + 1] - self.y[i1]) == sens:
                    entries["dir"] = "odd"
                else:
                    entries["dir"] = "even"
        else:
            sens = np.sign(self.x[1] - self.x[0])
            for key, entries in self.segments.items():
                i1 = entries["mark_samples"][0]
                if np.sign(self.x[i1 + 1] - self.x[i1]) == sens:
                    entries["dir"] = "odd"
                else:
                    entries["dir"] = "even"

    def write_stn(
        self, file, data1, x, y, *, data2=None, day=None, month=None, year=None
    ):
        """
        Writes Geometrics magnetic gradiometer file in ASCII.

        Parameters
        ----------
        Obligatory parameters:

        file : str
            File where to store the data.
        data1 : numpy float array [number_of_samples_per_line, number_of_lines]
            Data of sensor 1.
        x : numpy float array [number_of_samples_per_line]
            X coordinates of all measured points.
        y : numpy float array [number_of_samples_per_line, number_of_lines]
            Y coordinates of all measured points.

        Optional parameters:

        data2 : numpy float array [number_of_samples_per_line, number_of_lines]
            Data of sensor 2. Optional. Default: np.zeros_like(data1)
        day : int
            Day of acquisition. Optional. Default: self.day[0]
        month : int
            Month of acquisition. Optional. Default: self.month[0]
        year : int
            Year of acquisition (2 ciphers). Optional. Default: self.year[0]
        direction : int
            If direction == 0: measurements were done in Y direction,
            else in X direction. Optional. Default: self.direction

        Returns
        -------
        None.

        """
        try:
            if not data1:
                print("No data given, no Geometrics stn file written")
                return
        except ValueError:
            pass
        try:
            if not data2:
                if self.grad_flag:
                    data2 = np.zeros_like(data1)
        except ValueError:
            pass
        if not day:
            day = self.day[0]
        if not month:
            month = self.month[0]
        if not year:
            year = self.year[0]
        with open(file, "w", encoding="utf-8") as fh:
            fh.write("99 0 0\n")
            mark = 1
            izero = 0
            midmark = 4
            endline = 5
            startline = 36
            # Store gridded data
            if isinstance(data1, np.ndarray):
                try:
                    if not x or not y:
                        print("No coordinates given, no Geometrics stn file written")
                        return
                except ValueError:
                    pass
                ny, nx = data1.shape
                # Calculate ficticious acquisition times:
                #    Times start at 0 seconds of the day of acquisition of the first sample and
                #    increase by 0.1s between samples. Every line starts at the next full minute
                #    after the end of the former line.
                dt = ny * 0.1
                tref = np.arange(ny) * 0.1
                start = np.arange(nx) * np.ceil(dt)
                for ix, xx in enumerate(x):
                    t = start[ix] + tref
                    h = np.array(t / 3600.0, dtype=int)
                    tt = t - h * 3600
                    m = np.array(tt / 60.0, dtype=int)
                    s = tt - m * 60
                    nl = len(t)
                    fh.write(
                        f"3{xx:13.3f}{y[-1]:13.3f} {h[-1]:02d}:"
                        + f"{m[-1]:02d}:{s[-1]:05.2f} {month:02d}/"
                        + f"{day:02d}/{year:02d} {izero:11d}"
                        + f"{ix:12d}{mark:12d}{endline:5d}\n"
                    )
                    if self.grad_flag:
                        for iy in range(ny - 1, -1, -1):
                            fh.write(
                                f"0 {data1[iy,ix]:11.3f}{data2[iy,ix]:11.3f} "
                                + f"{h[iy]:02d}:{m[iy]:02d}:"
                                + f"{s[iy]:05.2f} {month:02d}/"
                                + f"{day:02d}/{year:02d}{izero:4d}\n"
                            )
                    else:
                        for iy in range(ny - 1, -1, -1):
                            fh.write(
                                f"0 {data1[iy,ix]:11.3f} "
                                + f"{h[iy]:02d}:{m[iy]:02d}:"
                                + f"{s[iy]:05.2f} {month:02d}/"
                                + f"{day:02d}/{year:02d}{izero:4d}\n"
                            )
                    fh.write(
                        f"3{xx:13.3f}{y[0]:13.3f} {h[0]:02d}:"
                        + f"{m[0]:02d}:{s[0]:05.2f} {month:02d}/"
                        + f"{day:02d}/{year:02d}{ny:11d}{ix:12d}"
                        + f"{izero:12d}{startline:5d}\n"
                    )
            # Store non-gridded data
            else:
                keys = list(data1.keys())
                nkeys = 0
                for nkeys, k in enumerate(keys):
                    if isinstance(k, str):
                        break
                for nl in range(nkeys - 1, -1, -1):
                    key = keys[nl]
                    marks = self.segments[key]["mark_samples"]
                    x = data1[key]["x"]
                    y = data1[key]["y"]
                    d1 = data1[key]["s1"]
                    if self.grad_flag:
                        d2 = data1[key]["s2"]
                    year = self.year[marks[0] : marks[-1]]
                    month = self.month[marks[0] : marks[-1]]
                    day = self.day[marks[0] : marks[-1]]
                    h = self.hour[marks[0] : marks[-1]]
                    m = self.minute[marks[0] : marks[-1]]
                    s = self.second[marks[0] : marks[-1]]
                    n1 = 0
                    n2 = 0
                    for i, j in enumerate(range(len(marks) - 1, 0, -1)):
                        if i == 0:
                            fh.write(
                                f"3{x[-1]:13.3f}{y[-1]:13.3f} {h[-1]:02d}:"
                                + f"{m[-1]:02d}:{s[-1]:05.2f} {month[-1]:02d}/"
                                + f"{day[-1]:02d}/{year[-1]:02d} {izero:11d}"
                                + f"{nl:12d}{j:12d}{endline:5d}\n"
                            )
                        else:
                            fh.write(
                                f"3{x[n1]:13.3f}{y[n1]:13.3f} {h[n1]:02d}:"
                                + f"{m[n1]:02d}:{s[n1]:05.2f} {month[n1]:02d}/"
                                + f"{day[n1]:02d}/{year[n1]:02d} {n2-n1:11d}"
                                + f"{nl:12d}{j:12d}{midmark:5d}\n"
                            )
                        n1 = marks[j - 1] - marks[0]
                        n2 = marks[j] - marks[0]
                        if self.grad_flag:
                            for iy in range(n2 - 1, n1 - 1, -1):
                                fh.write(
                                    f"0 {d1[iy]:11.3f}{d2[iy]:11.3f} "
                                    + f"{h[iy]:02d}:{m[iy]:02d}:"
                                    + f"{s[iy]:05.2f} {month[iy]:02d}/"
                                    + f"{day[iy]:02d}/{year[iy]:02d}{izero:4d}\n"
                                )
                        else:
                            for iy in range(ny - 1, n1 - 1, -1):
                                fh.write(
                                    f"0 {d1[iy]:11.3f} "
                                    + f"{h[iy]:02d}:{m[iy]:02d}:"
                                    + f"{s[iy]:05.2f} {month[iy]:02d}/"
                                    + f"{day[iy]:02d}/{year[iy]:02d}{izero:4d}\n"
                                )
                    fh.write(
                        f"3 {x[0]:13.3f}{y[0]:13.3f} {h[0]:02d}:"
                        + f"{m[0]:02d}:{s[0]:05.2f} {month[0]:02d}/"
                        + f"{day[0]:02d}/{year[0]:02d}{n2-n1:11d}{nl:12d}"
                        + f"{izero:12d}{startline:5d}\n"
                    )
        print(f"\nfile {file} written")

    def clean_data(
        self, min_fix=None, max_fix=None, percent_down=None, percent_up=None
    ):
        """
        Set data to np.nan under certain conditions which may be:

        Parameters
        ----------
        min_fix : float. Optional (default: None)
            All data below this value are set to nan
        max_fix : float. Optional (default: None)
            All data above this value are set to nan
        percent_down : float. Optional (default: None)
            The lowermost percentile values are set to nan. A value of 0.01 means
            that all values lower than the 1% quantile are set to None
        percent_up : float. Optional (default: None)
            The uppermost percentile values are set to nan. A value of 0.01 means
            that all values higher than the 99% quantile are set to None

        The operation is done in situ.
        """
        print("\n")
        if min_fix:
            self.sensor1[self.sensor1 < min_fix] = np.nan
            if self.grad_data:
                self.sensor2[self.sensor2 < min_fix] = np.nan
            print(f"Clip below {np.round(min_fix,1)}")
        if max_fix:
            self.sensor1[self.sensor1 > max_fix] = np.nan
            if self.grad_data:
                self.sensor2[self.sensor2 > max_fix] = np.nan
            print(f"Clip above {np.round(max_fix,1)}")
        if percent_down:
            vmin1 = np.nanquantile(self.sensor1, percent_down)
            self.sensor1[self.sensor1 < vmin1] = np.nan
            if self.grad_data:
                vmin2 = np.nanquantile(self.sensor2, percent_down)
                self.sensor2[self.sensor2 < vmin2] = np.nan
                print(
                    f"Clip below {np.round(vmin1,1)} for sensor 1 and "
                    + f"{np.round(vmin2,1)} for sensor 2"
                )
            else:
                print(f"Clip below {np.round(vmin1,1)}")
        if percent_up:
            vmax1 = np.nanquantile(self.sensor1, 1 - percent_up)
            self.sensor1[self.sensor1 > vmax1] = np.nan
            if self.grad_data:
                vmax2 = np.nanquantile(self.sensor2, 1 - percent_up)
                self.sensor2[self.sensor2 > vmax2] = np.nan
                print(
                    f"Clip above {np.round(vmax1,1)} for sensor 1 and "
                    + f"{np.round(vmax2,1)} for sensor 2"
                )
            else:
                print(f"Clip above {np.round(vmax1,1)}")
        if self.grad_data:
            self.grad = (self.sensor1 - self.sensor2) / self.d_sensor

    def line_days(self):
        """
        Create a dictionary containing for every day of data acquisition the
        numbers of the lines having been measured.

        Returns
        -------
        self.lines_per_day : dict
            Dictionary has as key the number of the day. It is implicitely
            supposed that the days are unique, i.e. measurements were not done
            over a period longer than a month. For every entrance, a list contains
            the line numbers

        """
        days = np.unique(self.day)
        for d in days:
            self.lines_per_day[d] = []
            for key, entries in self.segments.items():
                if self.day[entries["mark_samples"][0]] == d:
                    self.lines_per_day[d].append(key)

    def get_line(self, i_line=0):
        """
        Get data of one single line

        Parameters
        ----------
        i_line : int, optional
            Number of line to be extracted (counting starts at 0). The default is 0.

        Returns
        -------
        sensor1: numpy float array
            Data of sensor 1.
        sensor2: numpy float array
            Data of sensor 2.
        x: numpy float array
            X-coordinate of all data points extracted.
        y: numpy float array
            Y-coordinate of all data points extracted.
        year: numpy int array
            year of acquisition of all data points extracted.
        month: numpy int array
            month of acquisition of all data points extracted.
        day: numpy int array
            day of acquisition of all data points extracted.
        hour: numpy int array
            hour of acquisition of all data points extracted.
        minute: numpy int array
            minute of acquisition of all data points extracted.
        second: numpy float array
            second of acquisition of all data points extracted.
        time: numpy int array
            second of day of acquisition of all data points extracted.

        """
        n1 = self.segments[i_line]["mark_samples"][0]
        n2 = self.segments[i_line]["mark_samples"][-1]
        if self.grad_data:
            return (
                self.sensor1[n1:n2],
                self.sensor2[n1:n2],
                self.x[n1:n2],
                self.y[n1:n2],
                self.year[n1:n2],
                self.month[n1:n2],
                self.day[n1:n2],
                self.hour[n1:n2],
                self.minute[n1:n2],
                self.second[n1:n2],
                self.time[n1:n2],
            )
        dum = np.array([0])
        return (
            self.sensor1[n1:n2],
            dum,
            self.x[n1:n2],
            self.y[n1:n2],
            self.year[n1:n2],
            self.month[n1:n2],
            self.day[n1:n2],
            self.hour[n1:n2],
            self.minute[n1:n2],
            self.second[n1:n2],
            self.time[n1:n2],
        )

    def geometrics_lines(self):
        """
        Put all data into a simplified dictionary, one entry per line

        No input parameters

        Returns
        -------
        data: dictionary with one entry per line, key = number of line.
              Each entry is itself a dictionary containing the following entries:

              - "s1": Numpy float array with data of sensor 1
              - "s2": Numpy float array with data of sensor 2
                If only data of one single sensor were measured, "s2" contains only one zero.
              - "x":  E-W coordinates of data points
              - "y"   N-S coordinates of data points

        """
        data = {}
        for i in range(self.n_lines):
            s1, s2, x, y, _, _, _, _, _, _, _ = self.get_line(i)
            data[i] = {}
            data[i]["s1"] = s1
            data[i]["s2"] = s2
            data[i]["x"] = x
            data[i]["y"] = y
            return data

    def get_segment(self, i_line=0, i_seg=0):
        """
        Get data of one segment (data between two marks) of a line

        Parameters
        ----------
        i_line : int, optional
            Number of line to be extracted (counting starts at 0). The default is 0.
        i_seg : int, optional
            Number of segment within line to be extracted (counting starts at 0).
            The default is 0.

        Returns
        -------
        sensor1: numpy float array
            Data of sensor 1.
        sensor2: numpy float array
            Data of sensor 2.
        x: numpy float array
            X-coordinate of all data points extracted.
        y: numpy float array
            Y-coordinate of all data points extracted.
        year: numpy int array
            year of acquisition of all data points extracted.
        month: numpy int array
            month of acquisition of all data points extracted.
        day: numpy int array
            day of acquisition of all data points extracted.
        hour: numpy int array
            hour of acquisition of all data points extracted.
        minute: numpy int array
            minute of acquisition of all data points extracted.
        second: numpy float array
            second of acquisition of all data points extracted.
        time: numpy int array
            second of day of acquisition of all data points extracted.

        """

        n1 = self.segments[i_line]["mark_samples"][i_seg]
        n2 = self.segments[i_line]["mark_samples"][i_seg + 1]
        if self.grad_data:
            return (
                self.sensor1[n1:n2],
                self.sensor2[n1:n2],
                self.x[n1:n2],
                self.y[n1:n2],
                self.year[n1:n2],
                self.month[n1:n2],
                self.day[n1:n2],
                self.hour[n1:n2],
                self.minute[n1:n2],
                self.second[n1:n2],
                self.time[n1:n2],
            )
        dum = np.array([0])
        return (
            self.sensor1[n1:n2],
            dum,
            self.x[n1:n2],
            self.y[n1:n2],
            self.year[n1:n2],
            self.month[n1:n2],
            self.day[n1:n2],
            self.hour[n1:n2],
            self.minute[n1:n2],
            self.second[n1:n2],
            self.time[n1:n2],
        )

    def read_base(self, infile, year=None):
        """
        Read data from a base station. It is supposed the base station is a
        Geometrics PPM station. The function recognizes automatically whether
        it is a G-856 (old version) or a G-857 (newer version) instrument

        If other data have already been read in, the new ones are concatenated
        to the existing data. Before leaving the subroutine, data are sorted by
        measurement time. In this way, the order of data storage within a file
        and the order of reading of different files are without importance.

        Parameters
        ----------
        infile : str
            Name of file containing base station data.
        year : int, optional
            If base station data were recorded with a G-856 PPM, the date is
            coded as Julian day without the year of acquisition. In this case,
            the year should be given. If Geometrics G-858 (Cs magnetometer) data
            have been input already, the year of those data may be used and the
            variable may be left "None". The default is None.

        Data are stored in the following arrays (usually, they are base station data):

        - self.base : 1D numpy float array contains all base station data
        - self.day_base   : 1D numpy int array contains day of acquisition for all data
        - self.month_base : 1D numpy int array contains month of acquisition for all data
        - self.year_base  : 1D numpy int array contains year of acquisition for all data
        - self.hour_base  : 1D numpy int array contains hour of acquisition for all data
        - self.minute_base: 1D numpy int array contains minute of acquisition for all data
        - self.second_base: 1D numpy float array contains second of acquisition for all data
        - self.time_base  : 1D numpy float array contains second within year for all data

        """

        if not year:
            year = self.year[0]

        with open(infile, "r", encoding="utf-8") as fh:
            lines = fh.readlines()
        nlines = len(lines)
        self.year_base = list(self.year_base)
        self.month_base = list(self.month_base)
        self.day_base = list(self.day_base)
        self.jday_base = list(self.jday_base)
        self.hour_base = list(self.hour_base)
        self.minute_base = list(self.minute_base)
        self.second_base = list(self.second_base)
        self.base = list(self.base)
        self.time_base = list(self.time_base)
        if lines[0][0] == "*":
            for k in range(nlines):
                line = lines[k].split()
                try:
                    j_day = int(line[2])
                except IndexError:
                    break
                d, m = self.u.julian2date(j_day, year)
                self.jday_base.append(j_day)
                self.year_base.append(year)
                self.month_base.append(m)
                self.day_base.append(d)
                self.hour_base.append(int(line[3][0:2]))
                self.minute_base.append(int(line[3][2:4]))
                self.second_base.append(float(line[3][4:6]))
                self.time_base.append(j_day)
                if len(line) == 5:
                    self.base.append(float(line[-1].split("?")[1]) / 10.0)
                else:
                    self.base.append(float(line[-1]) / 10.0)
        else:
            for k in range(2, nlines):
                line = lines[k].split(",")
                try:
                    self.month_base.append(int(line[2][0:2]))
                except IndexError:
                    break
                self.day_base.append(int(line[2][3:5]))
                self.year_base.append(int(line[2][6:8]))
                self.hour_base.append(int(line[3][0:2]))
                self.jday_base.append(
                    self.u.date2julian(
                        self.day_base[-1], self.month_base[-1], self.year_base[-1]
                    )
                )
                self.minute_base.append(int(line[3][3:5]))
                self.second_base.append(float(line[3][6:8]))
                self.time_base.append(self.jday_base[-1])
                self.base.append(float(line[6]))
        self.month_base = np.array(self.month_base, dtype=int)
        self.day_base = np.array(self.day_base, dtype=int)
        self.year_base = np.array(self.year_base, dtype=int)
        self.hour_base = np.array(self.hour_base, dtype=int)
        self.minute_base = np.array(self.minute_base, dtype=int)
        self.second_base = np.array(self.second_base, dtype=float)
        self.base = np.array(self.base, dtype=float)
        self.time_base = (
            np.array(self.time_base) * 86400.0
            + (self.hour_base * 60.0 + self.minute_base) * 60.0
            + self.second_base
        )
        index = np.argsort(self.time_base)
        self.month_base = self.month_base[index]
        self.day_base = self.day_base[index]
        self.year_base = self.year_base[index]
        self.hour_base = self.hour_base[index]
        self.minute_base = self.minute_base[index]
        self.second_base = self.second_base[index]
        self.base = self.base[index]
        self.time_base = self.time_base[index]

    def write_base(self, outfile):
        """
        Write base station data in G-856 format.

        Parameters
        ----------
        outfile : str
            Name of file where to write the base station data.

        """
        with open(outfile, "w", encoding="utf-8") as fo:
            for i, b in enumerate(self.base):
                if np.isnan(b):
                    continue
                fo.write(
                    f"*  0 {self.jday_base[i]:3d} {self.hour_base[i]:02d}"
                    + f"{self.minute_base[i]:02d}{int(self.second_base[i]):02d}"
                    + f"{i:5d}{int(b*10):7d}\n"
                )
